import pytest

import numpy as np
import scipy.stats

from psyneulink.core.compositions import Composition
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism
from psyneulink.core.components.mechanisms.processing.objectivemechanism import ObjectiveMechanism
from psyneulink.core.globals.sampleiterator import SampleSpec
from psyneulink.core.components.mechanisms.modulatory.control.optimizationcontrolmechanism import OptimizationControlMechanism
from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
from psyneulink.core.globals.keywords import OVERRIDE
from psyneulink.core.components.functions.optimizationfunctions import ParamEstimationFunction

def test_moving_average():

    # Set an arbitrary seed and a global random state to keep the randomly generated quantities the same between runs
    seed = 20170530  # this will be separately given to ELFI
    np.random.seed(seed)

    # true parameters
    t1_true = 0.6
    t2_true = 0.2

    # Define a function that simulates a 2nd order moving average, assuming mean zero:
    # y_t = w_t + t1*w_t-1 + t2*w_t-2
    # where t1 and t2 are real and w_k is i.i.d sequence white noise with N(0,1)
    def MA2(input=[0], t1=0.5, t2=0.5, n_obs=100, batch_size=1, random_state=None):
        # FIXME: Convert arguments to scalar if they are not. Why is this nescessary?
        # PsyNeuLink, when creating a user defined function, seems to expect the function
        # to support inputs of type np.ndarray even when they are only allowed to be
        # scalars.
        n_obs = n_obs[0] if (type(n_obs) is np.ndarray) else n_obs
        batch_size = batch_size[0] if (type(batch_size) is np.ndarray) else batch_size

        # Make inputs 2d arrays for numpy broadcasting with w
        t1 = np.asanyarray(t1).reshape((-1, 1))
        t2 = np.asanyarray(t2).reshape((-1, 1))
        random_state = random_state or np.random

        w = random_state.randn(int(batch_size), int(n_obs) + 2)  # i.i.d. sequence ~ N(0,1)
        x = w[:, 2:] + t1 * w[:, 1:-1] + t2 * w[:, :-2]
        return x

    # Lets make some observed data. This will be the data we try to fit parameters for.
    y_obs = MA2(t1_true, t2_true)

    # Make a processing mechanism out of our simulator.
    ma_mech = ProcessingMechanism(function=MA2,
                                      size=1,
                                      name='Moving Average (2nd Order)')

    # Now lets add it to a composition
    comp = Composition(name="Moving_Average")
    comp.add_node(ma_mech)

    # Now lets setup some control signals for the parameters we want to
    # infer. This is where we would like to specify priors.
    signalSearchRange = SampleSpec(start=0.1, stop=2.0, step=0.2)
    t1_control_signal = ControlSignal(projections=[('t1', ma_mech)],
                                      allocation_samples=signalSearchRange,
                                      modulation=OVERRIDE)
    t2_control_signal = ControlSignal(projections=[('t2', ma_mech)],
                                      allocation_samples=signalSearchRange,
                                      modulation=OVERRIDE)

    # A function to calculate the auto-covariance with specific lag for a
    # time series. We will use this function to compute the summary statistics
    # for generated and observed data so that we can compute a metric between the
    # two. In PsyNeuLink terms, this will be part of an ObjectiveMechanism.
    # A function to calculate the auto-covariance with specific lag for a
    # time series. We will use this function to compute the summary statistics
    # for generated and observed data so that we can compute a metric between the
    # two. In PsyNeuLink terms, this will be part of an ObjectiveMechanism.
    def autocov(agent_rep, x=None, lag=1):
        if x is None:
            return np.asarray(0.0)

        C = np.mean(x[:, lag:] * x[:, :-lag], axis=1)
        return C

    # # Lets make one function that computes all the summary stats in one go because PsyNeuLink
    # # objective mechanism expect a single function.
    # def objective_function(x):
    #     return np.concatenate((autocov(x), autocov(x, lag=2)))
    #
    # # Objective Mechanism and its function currently need to be specified in the script.
    # # (In future versions, this will be set up automatically)
    # objective_mech = ObjectiveMechanism(function=objective_function,
    #                                         monitor=[ma_mech])

    # Setup the controller with the ParamEstimationFunction
    comp.add_controller(
        controller=OptimizationControlMechanism(
            agent_rep=comp,
            function=ParamEstimationFunction(
                priors={'t1': (scipy.stats.uniform, 0, 2), 't2': (scipy.stats.uniform, 0, 2)},
                observed=y_obs,
                summary=[(autocov, 1), (autocov, 2)],
                discrepancy='euclidean',
                n_samples=3, quantile=0.01, # Set very small now cause things are slow.
                seed=seed),
            objective_mechanism=False,
            control_signals=[t1_control_signal, t2_control_signal]))

    comp.disable_all_history()

    # Lets setup some input to the mechanism, not that it uses it for anything.
    stim_list_dict = {ma_mech: [0]}

    # # FIXME: Show graph fails when the controller doesn't have an objective mechanism.
    # comp.show_graph(show_controller=True,
    #                 show_projection_labels=True,
    #                 show_node_structure=True,
    #                 show_cim=True,
    #                 show_dimensions=True)

    comp.run(inputs=stim_list_dict)

    # FIXME: The final test should be to check if the true parameters set above are
    # recovered approximately. Not sure how to get all the samples out from
    # above yet though so just pass for now.
    assert True



import numpy as np
import pytest

import psyneulink as pnl
import psyneulink.core.llvm as pnlvm

class TestRearrange:

    @pytest.mark.function
    @pytest.mark.combination_function
    def test_no_default_variable(self):
        R_function = pnl.Rearrange(arrangement=[(1,2),0])
        result = R_function.execute([[0,0],[1,1],[2,2]])
        for exp,act in zip(result, [[ 1.,  1.,  2.,  2.],[ 0.,  0.]]):
            assert np.allclose(exp,act)

    @pytest.mark.function
    @pytest.mark.combination_function
    def test_with_default_variable(self):
        R_function = pnl.Rearrange(default_variable=[[0],[0],[0]], arrangement=[(1,2),0])
        result = R_function.execute([[0,0],[1,1],[2,2]])
        for exp,act in zip(result, [[ 1.,  1.,  2.,  2.],[ 0.,  0.]]):
            assert np.allclose(exp,act)

    @pytest.mark.function
    @pytest.mark.combination_function
    def test_arrangement_has_out_of_bounds_index(self):
        with pytest.raises(pnl.FunctionError) as error_text:
            pnl.Rearrange(default_variable=[0,0], arrangement=[(1,2),0])
        error_msg = "'default_variable' for Rearrange must be at least 2d."
        assert error_msg in str(error_text.value)

    @pytest.mark.function
    @pytest.mark.combination_function
    def test_default_variable_mismatches_arrangement(self):
        with pytest.raises(pnl.FunctionError) as error_text:
            pnl.Rearrange(default_variable=[[0],[0]], arrangement=[(1,2),0])
        error_msg_a = "'arrangement' arg for Rearrange"
        error_msg_b = "is out of bounds for its 'default_variable' arg (max index = 1)."
        assert error_msg_a in str(error_text.value)
        assert error_msg_b in str(error_text.value)

    @pytest.mark.function
    @pytest.mark.combination_function
    def test_default_variable_has_non_numeric_index(self):
        # with pytest.raises(pnl.FunctionError) as error_text:
        with pytest.raises(pnl.FunctionError) as error_text:
            pnl.Rearrange(default_variable=[[0],['a']], arrangement=[(1,2),0])
        # error_msg = "All elements of 'default_variable' for Rearrange must be scalar values."
        error_msg = "must be scalar values"
        assert error_msg in str(error_text.value)

    @pytest.mark.function
    @pytest.mark.combination_function
    def test_arrangement_has_non_numeric_index(self):
        with pytest.raises(pnl.FunctionError) as error_text:
            pnl.Rearrange(default_variable=[[0],[0],[0]], arrangement=[(1,2),'a'])
        error_msg_a = "Index specified in 'arrangement' arg"
        error_msg_b = "('a') is not an int."
        assert error_msg_a in str(error_text.value)
        assert error_msg_b in str(error_text.value)

    # @pytest.mark.function
    # @pytest.mark.combination_function
    # def test_column_vector(self):
    #     R_function = pnl.core.components.functions.combinationfunctions.Reduce(operation=pnl.SUM)
    #     R_mechanism = pnl.ProcessingMechanism(function=pnl.core.components.functions.combinationfunctions.Reduce(operation=pnl.SUM),
    #                                           default_variable=[[1], [2], [3], [4], [5]],
    #                                           name="R_mechanism")
    #
    #     assert np.allclose(R_function.execute([[1], [2], [3], [4], [5]]), [1, 2, 3, 4, 5])
    #     # assert np.allclose(R_function.execute([[1], [2], [3], [4], [5]]), [15.0])
    #     assert np.allclose(R_function.execute([[[1], [2], [3], [4], [5]]]), [15.0])
    #
    #     assert np.allclose(R_mechanism.execute([[1], [2], [3], [4], [5]]), [1, 2, 3, 4, 5])
    #     # assert np.allclose(R_mechanism.execute([[1], [2], [3], [4], [5]]), [15.0])
    #
    # @pytest.mark.function
    # @pytest.mark.combination_function
    # def test_matrix(self):
    #     R_function = pnl.core.components.functions.combinationfunctions.Reduce(operation=pnl.SUM)
    #     R_mechanism = pnl.ProcessingMechanism(function=pnl.core.components.functions.combinationfunctions.Reduce(operation=pnl.SUM),
    #                                           default_variable=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
    #                                           name="R_mechanism")
    #
    #     assert np.allclose(R_function.execute([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), [6, 15, 24])
    #     assert np.allclose(R_function.execute([[[1, 2, 3], [4, 5, 6], [7, 8, 9]]]), [12, 15, 18])
    #
    #     assert np.allclose(R_mechanism.execute([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), [6, 15, 24])


class TestReduce:

    @pytest.mark.function
    @pytest.mark.combination_function
    def test_single_array(self):
        R_function = pnl.Reduce(operation=pnl.SUM)
        R_mechanism = pnl.ProcessingMechanism(function=pnl.Reduce(operation=pnl.SUM),
                                              default_variable=[[1, 2, 3, 4, 5]],
                                              name="R_mechanism")

        assert np.allclose(R_function.execute([1, 2, 3, 4, 5]), [15.0])
        assert np.allclose(R_function.execute([[1, 2, 3, 4, 5]]), [15.0])
        assert np.allclose(R_function.execute([[[1, 2, 3, 4, 5]]]), [1, 2, 3, 4, 5])
        # assert np.allclose(R_function.execute([[[1, 2, 3, 4, 5]]]), [15.0])

        assert np.allclose(R_mechanism.execute([1, 2, 3, 4, 5]), [[15.0]])
        assert np.allclose(R_mechanism.execute([[1, 2, 3, 4, 5]]), [[15.0]])
        assert np.allclose(R_mechanism.execute([1, 2, 3, 4, 5]), [15.0])
        # assert np.allclose(R_mechanism.execute([[1, 2, 3, 4, 5]]), [15.0])

    @pytest.mark.function
    @pytest.mark.combination_function
    def test_column_vector(self):
        R_function = pnl.Reduce(operation=pnl.SUM)
        R_mechanism = pnl.ProcessingMechanism(function=pnl.Reduce(operation=pnl.SUM),
                                              default_variable=[[1], [2], [3], [4], [5]],
                                              name="R_mechanism")

        assert np.allclose(R_function.execute([[1], [2], [3], [4], [5]]), [1, 2, 3, 4, 5])
        # assert np.allclose(R_function.execute([[1], [2], [3], [4], [5]]), [15.0])
        assert np.allclose(R_function.execute([[[1], [2], [3], [4], [5]]]), [15.0])

        assert np.allclose(R_mechanism.execute([[1], [2], [3], [4], [5]]), [1, 2, 3, 4, 5])
        # assert np.allclose(R_mechanism.execute([[1], [2], [3], [4], [5]]), [15.0])

    @pytest.mark.function
    @pytest.mark.combination_function
    def test_matrix(self):
        R_function = pnl.Reduce(operation=pnl.SUM)
        R_mechanism = pnl.ProcessingMechanism(function=pnl.Reduce(operation=pnl.SUM),
                                              default_variable=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                                              name="R_mechanism")

        assert np.allclose(R_function.execute([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), [6, 15, 24])
        assert np.allclose(R_function.execute([[[1, 2, 3], [4, 5, 6], [7, 8, 9]]]), [12, 15, 18])

        assert np.allclose(R_mechanism.execute([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), [6, 15, 24])

    # def test_heterogeneous_arrays(self):
    #     R_function = pnl.Reduce(operation=pnl.SUM)
    #     # R_mechanism = pnl.ProcessingMechanism(function=pnl.Reduce(operation=pnl.SUM),
    #     #                                       default_variable=[[1, 2], [3, 4, 5], [6, 7, 8, 9]],
    #     #                                       name="R_mechanism")
    #     print(R_function.execute([[1, 2], [3, 4, 5], [6, 7, 8, 9]]))
    #     print(R_function.execute([[[1, 2], [3, 4, 5], [6, 7, 8, 9]]]))
    #
    #     # print("mech = ", R_mechanism.execute([[1, 2], [3, 4, 5], [6, 7, 8, 9]]))
    #     # print("mech = ", R_mechanism.execute([[[1, 2], [3, 4, 5], [6, 7, 8, 9]]]))
    #     # print("mech = ", R_mechanism.execute([[[1, 2], [3, 4, 5], [6, 7, 8, 9]]]))
    #

SIZE=5
np.random.seed(0)
#This gives us the correct 2d array
test_varr1 = np.random.rand(1, SIZE)
test_varr2 = np.random.rand(2, SIZE)
test_varr3 = np.random.rand(3, SIZE)

#This gives us the correct 2d column array
test_varc1 = np.random.rand(SIZE, 1)
test_varc2 = np.random.rand(SIZE, 1)
test_varc3 = np.random.rand(SIZE, 1)

#This gives us the correct 2d matrix array
test_varm1 = np.random.rand(SIZE, 3)
test_varm2 = np.random.rand(SIZE, 3)
test_varm3 = np.random.rand(SIZE, 3)

RAND1_V = np.random.rand(SIZE)
RAND2_V = np.random.rand(SIZE)
RAND3_V = np.random.rand(SIZE)

RAND1_S = np.random.rand()
RAND2_S = np.random.rand()
RAND3_S = np.random.rand()


@pytest.mark.benchmark(group="ReduceFunction")
@pytest.mark.function
@pytest.mark.combination_function
@pytest.mark.parametrize("variable", [test_varr1, test_varr2, test_varr3,
                                      test_varc1, test_varc2, test_varc3,
                                      test_varm1, test_varm2, test_varm3,
                                     ], ids=["VAR1", "VAR2", "VAR3",
                                             "VAR1c", "VAR2c", "VAR3c",
                                             "VAR1m", "VAR2m", "VAR3m",
                                            ])
@pytest.mark.parametrize("operation", [pnl.SUM, pnl.PRODUCT])
@pytest.mark.parametrize("exponents", [None, 2.0, [3.0], 'V'], ids=["E_NONE", "E_SCALAR", "E_VECTOR1", "E_VECTORN"])
@pytest.mark.parametrize("weights", [None, 0.5, 'VC', 'VR'], ids=["W_NONE", "W_SCALAR", "W_VECTORN", "W_VECTORM"])
@pytest.mark.parametrize("scale", [RAND1_S, RAND1_V], ids=["S_SCALAR", "S_VECTOR"])
@pytest.mark.parametrize("offset", [RAND2_S, RAND2_V], ids=["O_SCALAR", "O_VECTOR"])
@pytest.mark.parametrize("mode", ['Python',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_reduce_function(variable, operation, exponents, weights, scale, offset, mode, benchmark):
    if weights == 'VC':
        weights = [[(-1) ** i] for i, v in enumerate(variable)]
    if weights == 'VR':
        weights = [(-1) ** i for i, v in enumerate(variable[0])]
    if exponents == 'V':
        exponents = [[v[0]] for v in variable]

    try:
        f = pnl.Reduce(default_variable=variable,
                       operation=operation,
                       exponents=exponents,
                       weights=weights,
                       scale=scale,
                       offset=offset)
    except ValueError as e:
        if not np.isscalar(scale) and "The truth value of an array" in str(e):
            pytest.xfail("vector scale is not supported")
        if not np.isscalar(offset) and "The truth value of an array" in str(e):
            pytest.xfail("vector offset is not supported")
        raise e from None

    if (mode == 'LLVM'):
        e = pnlvm.execution.FuncExecution(f)
        res = benchmark(e.execute, variable)
    elif (mode == 'PTX'):
        e = pnlvm.execution.FuncExecution(f)
        res = benchmark(e.cuda_execute, variable)
    else:
        res = benchmark(f.function, variable)

    scale = 1.0 if scale is None else scale
    offset = 0.0 if offset is None else offset
    exponent = 1.0 if exponents is None else exponents
    weights = 1.0 if weights is None else weights

    tmp = (variable ** exponent) * weights
    if operation == pnl.SUM:
        expected = np.sum(tmp, axis=1) * scale + offset
    if operation == pnl.PRODUCT:
        expected = np.product(tmp, axis=1) * scale + offset

    assert np.allclose(res, expected)


@pytest.mark.benchmark(group="LinearCombinationFunction")
@pytest.mark.function
@pytest.mark.combination_function
@pytest.mark.parametrize("variable", [test_varr1, test_varr2], ids=["VAR1", "VAR2"])
@pytest.mark.parametrize("operation", [pnl.SUM, pnl.PRODUCT])
@pytest.mark.parametrize("exponents", [None, 2.0, [3.0], 'V'], ids=["E_NONE", "E_SCALAR", "E_VECTOR1", "E_VECTORN"])
@pytest.mark.parametrize("weights", [None, 0.5, 'V'], ids=["W_NONE", "W_SCALAR", "W_VECTORN"])
@pytest.mark.parametrize("scale", [None, RAND1_S, RAND1_V], ids=["S_NONE", "S_SCALAR", "S_VECTOR"])
@pytest.mark.parametrize("offset", [None, RAND2_S, RAND2_V], ids=["O_NONE", "O_SCALAR", "O_VECTOR"])
@pytest.mark.parametrize("bin_execute", ['Python',
                                        pytest.param('LLVM', marks=pytest.mark.llvm),
                                        pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_linear_combination_function(variable, operation, exponents, weights, scale, offset, bin_execute, benchmark):
    if weights == 'V':
        weights = [[-1 ** i] for i, v in enumerate(variable)]
    if exponents == 'V':
        exponents = [[v[0]] for v in variable]

    f = pnl.LinearCombination(default_variable=variable,
                              operation=operation,
                              exponents=exponents,
                              weights=weights,
                              scale=scale,
                              offset=offset)
    if (bin_execute == 'LLVM'):
        e = pnlvm.execution.FuncExecution(f)
        res = benchmark(e.execute, variable)
    elif (bin_execute == 'PTX'):
        e = pnlvm.execution.FuncExecution(f)
        res = benchmark(e.cuda_execute, variable)
    else:
        res = benchmark(f.function, variable)

    scale = 1.0 if scale is None else scale
    offset = 0.0 if offset is None else offset
    exponent = 1.0 if exponents is None else exponents
    weights = 1.0 if weights is None else weights

    tmp = (variable ** exponent) * weights
    if operation == pnl.SUM:
        expected = np.sum(tmp, axis=0) * scale + offset
    if operation == pnl.PRODUCT:
        expected = np.product(tmp, axis=0) * scale + offset

    assert np.allclose(res, expected)


@pytest.mark.benchmark(group="LinearCombinationFunction in Mechanism")
@pytest.mark.function
@pytest.mark.combination_function
@pytest.mark.parametrize("operation", [pnl.SUM, pnl.PRODUCT])
@pytest.mark.parametrize("input, input_ports", [ ([[1,2,3,4]], ["hi"]), ([[1,2,3,4], [5,6,7,8], [9,10,11,12]], ['1','2','3']), ([[1, 2, 3, 4], [5, 6, 7, 8], [0, 0, 1, 2]], ['1','2','3']) ], ids=["1S", "2S", "3S"])
@pytest.mark.parametrize("scale", [None, 2.5, [1,2.5,0,0]], ids=["S_NONE", "S_SCALAR", "S_VECTOR"])
@pytest.mark.parametrize("offset", [None, 1.5, [1,2.5,0,0]], ids=["O_NONE", "O_SCALAR", "O_VECTOR"])
@pytest.mark.parametrize("mode", ['Python',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_linear_combination_function_in_mechanism(operation, input, input_ports, scale, offset, benchmark, mode):
    f = pnl.LinearCombination(default_variable=input, operation=operation, scale=scale, offset=offset)
    p = pnl.ProcessingMechanism(size=[len(input[0])] * len(input), function=f, input_ports=input_ports)

    if mode == 'Python':
        res = benchmark(f.execute, input)
    elif mode == 'LLVM':
        e = pnlvm.execution.FuncExecution(f)
        res = benchmark(e.execute, input)
    elif mode == 'PTX':
        e = pnlvm.execution.FuncExecution(f)
        res = benchmark(e.cuda_execute, input)


    scale = 1.0 if scale is None else scale
    offset = 0.0 if offset is None else offset
    if operation == pnl.SUM:
        expected = np.sum(input, axis=0) * scale + offset
    if operation == pnl.PRODUCT:
        expected = np.product(input, axis=0) * scale + offset

    assert np.allclose(res, expected)



import ctypes
import numpy as np
import pytest

from psyneulink.core import llvm as pnlvm

from llvmlite import ir


ITERATIONS=100
DIM_X=1000

matrix = np.random.rand(DIM_X, DIM_X)
vector = np.random.rand(DIM_X)
llvm_res = np.random.rand(DIM_X)

x, y = matrix.shape

@pytest.mark.llvm
@pytest.mark.parametrize('mode', ['CPU',
                                  pytest.param('PTX', marks=pytest.mark.cuda)])
def test_fixed_dimensions__pnl_builtin_vxm(mode):
    # The original builtin mxv function
    binf = pnlvm.LLVMBinaryFunction.get("__pnl_builtin_vxm")
    orig_res = np.empty_like(llvm_res)
    if mode == 'CPU':
        ct_in_ty, ct_mat_ty, _, _, ct_res_ty = binf.byref_arg_types

        ct_vec = vector.ctypes.data_as(ctypes.POINTER(ct_in_ty))
        ct_mat = matrix.ctypes.data_as(ctypes.POINTER(ct_mat_ty))
        ct_res = orig_res.ctypes.data_as(ctypes.POINTER(ct_res_ty))

        binf.c_func(ct_vec, ct_mat, x, y, ct_res)
    else:
        binf.cuda_wrap_call(vector, matrix, np.int32(x), np.int32(y), orig_res)

    custom_name = None

    with pnlvm.LLVMBuilderContext() as ctx:
        custom_name = ctx.get_unique_name("vxsqm")
        double_ptr_ty = ctx.convert_python_struct_to_llvm_ir(1.0).as_pointer()
        func_ty = ir.FunctionType(ir.VoidType(), (double_ptr_ty, double_ptr_ty, double_ptr_ty))

        # get builtin IR
        builtin = ctx.import_llvm_function("__pnl_builtin_vxm")

        # Create square vector matrix multiply
        function = ir.Function(ctx.module, func_ty, name=custom_name)
        _x = ctx.int32_ty(x)
        _v, _m, _o = function.args
        block = function.append_basic_block(name="entry")
        builder = ir.IRBuilder(block)
        builder.call(builtin, [_v, _m, _x, _x, _o])
        builder.ret_void()

    binf2 = pnlvm.LLVMBinaryFunction.get(custom_name)
    new_res = np.empty_like(llvm_res)

    if mode == 'CPU':
        ct_res = new_res.ctypes.data_as(ctypes.POINTER(ct_res_ty))

        binf2(ct_vec, ct_mat, ct_res)
    else:
        binf2.cuda_wrap_call(vector, matrix, new_res)

    assert np.array_equal(orig_res, new_res)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# **************************************  DefaultControlMechanism ************************************************

"""

The DefaultControlMechanism is created for a `System` if no other controller type is specified. The
DefaultControlMechanism creates an `ControlSignal` for each `ControlProjection` it is assigned, and uses
`defaultControlAllocation` as the `value <ControlSignal.value>` for the ControlSignal.  By default,
`defaultControlAllocation` =  1, so that ControlProjections from the DefaultControlMechanism have no effect on their
parameters.  However, it can be used to uniformly control the parameters that receive ControlProjections from it,
by manually changing the value of `defaultControlAllocation`.  See `ControlMechanism <ControlMechanism>` for additional
details of how ControlMechanism are created, executed and their attributes.

COMMENT:
   ADD LINK FOR defaultControlAllocation

    TEST FOR defaultControlAllocation:  |defaultControlAllocation|

    ANOTHER TEST FOR defaultControlAllocation:  :py:print:`defaultControlAllocation`

    AND YET ANOTHER TEST FOR defaultControlAllocation:  :py:print:|defaultControlAllocation|

    LINK TO DEFAULTS: :doc:`Defaults`
COMMENT


"""

import numpy as np
import typecheck as tc

from psyneulink.core.components.mechanisms.modulatory.control.controlmechanism import ControlMechanism
from psyneulink.core.components.mechanisms.processing.objectivemechanism import ObjectiveMechanism
from psyneulink.core.globals.defaults import defaultControlAllocation
from psyneulink.core.globals.keywords import CONTROL, INPUT_PORTS, NAME
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel
from psyneulink.core.globals.utilities import ContentAddressableList

__all__ = [
    'DefaultControlMechanism', 'DefaultControlMechanismError'
]


class DefaultControlMechanismError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value


class DefaultControlMechanism(ControlMechanism):
    """Subclass of `ControlMechanism <ControlMechanism>` that implements a DefaultControlMechanism.

    COMMENT:
        Description:
            Implements default source of control signals, with one inputPort and outputPort for each.
            Uses defaultControlAllocation as input(s) and pass value(s) unchanged to outputPort(s) and
            ControlProjection(s)

            Every ControlProjection is assigned this Mechanism as its sender by default (i.e., unless a sender is
                explicitly specified in its constructor).

            An inputPort and outputPort is created for each ControlProjection assigned:
                the inputPort is assigned the
                :py:constant:`defaultControlAllocation <Defaults.defaultControlAllocation>` value;
                when the DefaultControlMechanism executes, it simply assigns the same value to the ControlProjection.

            Class attributes:
                + componentType (str): System Default Mechanism
    COMMENT
    """

    componentType = "DefaultControlMechanism"

    classPreferenceLevel = PreferenceLevel.SUBTYPE
    # classPreferenceLevel = PreferenceLevel.TYPE

    # Any preferences specified below will override those specified in TYPE_DEFAULT_PREFERENCES
    # Note: only need to specify setting;  level will be assigned to Type automatically
    # classPreferences = {
    #     PREFERENCE_SET_NAME: 'DefaultControlMechanismCustomClassPreferences',
    #     PREFERENCE_KEYWORD<pref>: <setting>...}

    from psyneulink.core.components.functions.transferfunctions import Linear

    @tc.typecheck
    def __init__(self,
                 objective_mechanism:tc.optional(tc.any(ObjectiveMechanism, list))=None,
                 control_signals:tc.optional(list)=None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None,
                 function=None,
                 **kwargs
                 ):

        super(DefaultControlMechanism, self).__init__(
                objective_mechanism=objective_mechanism,
                control_signals=control_signals,
                function=function,
                params=params,
                name=name,
                prefs=prefs,

                **kwargs)

    def _instantiate_input_ports(self, context=None):
        """Instantiate input_value attribute

        Instantiate input_ports and monitored_output_ports attributes (in case they are referenced)
            and assign any OutputPorts that project to the input_ports to monitored_output_ports

        IMPLEMENTATION NOTE:  At present, these are dummy assignments, simply to satisfy the requirements for
                              subclasses of ControlMechanism;  in the future, an _instantiate_objective_mechanism()
                              method should be implemented that also implements an _instantiate_monitored_output_ports
                              method, and that can be used to add OutputPorts/Mechanisms to be monitored.
        """

        if not hasattr(self, INPUT_PORTS):
            self._input_ports = None
        elif self.input_ports:
            for input_port in self.input_ports:
                for projection in input_port.path_afferents:
                    self.monitored_output_ports.append(projection.sender)

    def _instantiate_control_signal(self, control_signal, context=None):
        """Instantiate requested ControlSignal, ControlProjection and associated InputPort
        """
        from psyneulink.core.components.ports.parameterport import ParameterPort

        if isinstance(control_signal, dict):
            if CONTROL in control_signal:
                projection = control_signal[CONTROL][0]
                input_name = 'DefaultControlAllocation for ' + projection.receiver.name + '_ControlSignal'
            elif NAME in control_signal:
                input_name = 'DefaultControlAllocation for ' + control_signal[NAME] + '_ControlSignal'

        elif isinstance(control_signal, tuple):
            input_name = 'DefaultControlAllocation for ' + control_signal[0] + '_ControlSignal'

        elif isinstance(control_signal, ParameterPort):
            input_name = 'DefaultControlAllocation for ' + control_signal.name + '_ControlSignal'

        else:
            raise DefaultControlMechanismError("control signal ({}) was not a dict, tuple, or ParameterPort".
                                               format(control_signal))

        # Instantiate input_ports and control_allocation attribute for control_signal allocations
        self._instantiate_default_input_port(input_name, [defaultControlAllocation], context=context)
        self.control_allocation = self.input_values

        # Call super to instantiate ControlSignal
        # Note: any params specified with ControlProjection for the control_signal
        #           should be in PARAMS entry of dict passed in control_signal arg
        control_signal = super()._instantiate_control_signal(control_signal=control_signal, context=context)

    def _instantiate_default_input_port(self, input_port_name, input_port_value, context=None):
        """Instantiate inputPort for ControlMechanism

        NOTE: This parallels ObjectMechanism._instantiate_input_port_for_monitored_port()
              It is implemented here to spare having to instantiate a "dummy" (and superfluous) ObjectiveMechanism
              for the sole purpose of creating input_ports for each value of defaultControlAllocation to assign
              to the ControlProjections.

        Extend self.defaults.variable by one item to accommodate new inputPort
        Instantiate the inputPort using input_port_name and input_port_value
        Update self.input_port and self.input_ports

        Args:
            input_port_name (str):
            input_port_value (2D np.array):
            context:

        Returns:
            input_port (InputPort):

        """

        # First, test for initialization conditions:

        # This is for generality (in case, for any subclass in the future, variable is assigned to None on init)
        if self.defaults.variable is None:
            self.defaults.variable = np.atleast_2d(input_port_value)

        # If there is a single item in self.defaults.variable, it could be the one assigned on initialization
        #     (in order to validate ``function`` and get its return value as a template for value);
        #     in that case, there should be no input_ports yet, so pass
        #     (i.e., don't bother to extend self.defaults.variable): it will be used for the new inputPort
        elif len(self.defaults.variable) == 1:
            if self.input_ports:
                self.defaults.variable = np.append(self.defaults.variable, np.atleast_2d(input_port_value), 0)
            else:
                # If there are no input_ports, this is the usual initialization condition;
                # Pass to create a new inputPort that will be assigned to existing the first item of self.defaults.variable
                pass
        # Other than on initialization (handled above), it is a PROGRAM ERROR if
        #    the number of input_ports is not equal to the number of items in self.defaults.variable
        elif len(self.defaults.variable) != len(self.input_ports):
            raise DefaultControlMechanismError(
                "PROGRAM ERROR:  The number of input_ports ({}) does not match "
                "the number of items found for the variable attribute ({}) of {}"
                "when creating {}".format(
                    len(self.input_ports),
                    len(self.defaults.variable),
                    self.name,
                    input_port_name,
                )
            )

        # Extend self.defaults.variable to accommodate new inputPort
        else:
            self.defaults.variable = np.append(self.defaults.variable, np.atleast_2d(input_port_value), 0)

        variable_item_index = self.defaults.variable.size - 1

        # Instantiate inputPort
        from psyneulink.core.components.ports.port import _instantiate_port
        from psyneulink.core.components.ports.inputport import InputPort
        input_port = _instantiate_port(owner=self,
                                         port_type=InputPort,
                                         name=input_port_name,
                                         reference_value=np.array(self.defaults.variable[variable_item_index]),
                                         reference_value_name='Default control allocation',
                                         params=None,
                                         context=context)

        #  Update inputPort and input_ports
        if self.input_ports:
            self._input_ports[input_port.name] = input_port
        else:
            from psyneulink.core.components.ports.port import Port_Base
            self._input_ports = ContentAddressableList(component_type=Port_Base,
                                                        list=[input_port],
                                                        name=self.name + '.input_ports')

        # self.input_value = [port.value for port in self.input_ports]

        return input_port



from psyneulink import *

# This script implements the following network, first described in Rumelhart and Todd
# (Rumelhart, D. E., & Todd, P. M. (1993). Learning and connectionist representations. Attention and performance XIV:
#  Synergies in experimental psychology, artificial intelligence, and cognitive neuroscience, 3-30).

# Semantic Network:

#    (REP_OUT)  (PROP_OUT) (QUAL_OUT) (ACT_OUT)
#  Represention  Property    Quality   Action
#           \________\_________/________/
#                        |
#                 Relations_Hidden
#                   _____|_____
#                  /           \
#   Representation_Hidden  Relations_Input
#                /           (REL_IN)
#               /
#   Representation_Input (REP_IN)

# Construct Mechanisms
rep_in = TransferMechanism(size=10, name='REP_IN')
rel_in = TransferMechanism(size=11, name='REL_IN')
rep_hidden = TransferMechanism(size=4, function=Logistic, name='REP_HIDDEN')
rel_hidden = TransferMechanism(size=5, function=Logistic, name='REL_HIDDEN')
rep_out = TransferMechanism(size=10, function=Logistic, name='REP_OUT')
prop_out = TransferMechanism(size=12, function=Logistic, name='PROP_OUT')
qual_out = TransferMechanism(size=13, function=Logistic, name='QUAL_OUT')
act_out = TransferMechanism(size=14, function=Logistic, name='ACT_OUT')

# Construct Composition
comp = Composition(name='Rumelhart Semantic Network')
comp.add_backpropagation_learning_pathway(pathway=[rel_in, rel_hidden])
comp.add_backpropagation_learning_pathway(pathway=[rel_hidden, rep_out])
comp.add_backpropagation_learning_pathway(pathway=[rel_hidden, prop_out])
comp.add_backpropagation_learning_pathway(pathway=[rel_hidden, qual_out])
comp.add_backpropagation_learning_pathway(pathway=[rel_hidden, act_out])
comp.add_backpropagation_learning_pathway(pathway=[rep_in, rep_hidden, rel_hidden])
comp.show_graph(show_learning=True)



import inspect
import psyneulink as pnl
import pytest

from psyneulink.core.components.projections.modulatory.modulatoryprojection import ModulatoryProjection_Base
from psyneulink.core.components.projections.pathway.pathwayprojection import PathwayProjection_Base

# gather all Component classes (a set to ensure no duplicates)
component_classes = []
component_class_constructor_arguments = {}
for item in pnl.__all__:
    evaled = eval(f'pnl.{item}')

    if isinstance(
        evaled,
        pnl.core.components.component.ComponentsMeta
    ):
        component_classes.append(evaled)
        component_class_constructor_arguments[evaled] = inspect.signature(
            evaled.__init__
        ).parameters

component_classes.sort(key=lambda x: x.__name__)


@pytest.mark.parametrize(
    'class_type',
    [
        pnl.Mechanism_Base,
        pnl.Function_Base,
        pnl.Port_Base,
        ModulatoryProjection_Base,
        PathwayProjection_Base,
    ]
)
def test_abstract_classes(class_type):
    with pytest.raises(TypeError) as err:
        class_type()
    assert 'abstract class' in str(err.value)


@pytest.mark.parametrize(
    'class_',
    component_classes
)
def test_function_parameters_stateless(class_):
    try:
        assert class_.parameters.function.stateful is False, (
            f'{class_.__name__}.parameters.function.stateful is True. '
            'The function Parameter for Components is currently '
            'expected to be stateless (defined as stateful=False)'
        )
    except AttributeError:
        pass


@pytest.mark.parametrize(
    'class_',
    component_classes
)
def test_parameters_user_specified(class_):
    violators = set()
    constructor_parameters = inspect.signature(class_.__init__).parameters
    for name, param in constructor_parameters.items():
        if (
            param.kind in {
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.KEYWORD_ONLY
            }
            and name in class_.parameters.names()
            and param.default is not inspect.Parameter.empty
            and param.default is not None
        ):
            violators.add(name)

    message = (
        "If a value other than None is used as the default value in a class's"
        + ' constructor/__init__, for an argument corresponding to a Parameter,'
        + ' _user_specified will always be True. The default value should be'
        + " specified in the class's Parameters inner class. Violators for"
        + f' {class_.__name__}: {violators}'
    )
    assert violators == set(), message



import numpy as np
import pytest

import psyneulink.core.components.functions.statefulfunctions.memoryfunctions as Functions
import psyneulink.core.llvm as pnlvm
from psyneulink import *

np.random.seed(0)
SIZE=10
test_var = np.random.rand(2, SIZE)
test_initializer = np.array([[test_var[0], test_var[1]]])
test_noise_arr = np.random.rand(SIZE)

RAND1 = np.random.random(1)
RAND2 = np.random.random()

test_data = [
# Default initializer does not work
#    (Functions.Buffer, test_var, {'rate':RAND1}, [[0.0],[0.0]]),
    (Functions.Buffer, test_var[0], {'history':512, 'rate':RAND1, 'initializer':[test_var[0]]}, [[0.03841128, 0.05005587, 0.04218721, 0.0381362 , 0.02965146, 0.04520592, 0.03062659, 0.0624149 , 0.06744644, 0.02683695],[0.14519169, 0.18920736, 0.15946443, 0.1441519 , 0.11208025, 0.17087491, 0.11576615, 0.23592355, 0.25494239, 0.10144161]]),
    (Functions.ContentAddressableMemory, test_var, {'rate':RAND1}, [[
       0.5488135039273248, 0.7151893663724195, 0.6027633760716439, 0.5448831829968969, 0.4236547993389047, 0.6458941130666561, 0.4375872112626925, 0.8917730007820798, 0.9636627605010293, 0.3834415188257777], [
       0.7917250380826646, 0.5288949197529045, 0.5680445610939323, 0.925596638292661, 0.07103605819788694, 0.08712929970154071, 0.02021839744032572, 0.832619845547938, 0.7781567509498505, 0.8700121482468192 ]]),
    (Functions.ContentAddressableMemory, test_var, {'rate':RAND1, 'retrieval_prob':0.5},
       [[ 0. for i in range(SIZE) ],[ 0. for i in range(SIZE) ]]),
    (Functions.ContentAddressableMemory, test_var, {'rate':RAND1, 'storage_prob':0.1},
       [[ 0. for i in range(SIZE) ],[ 0. for i in range(SIZE) ]]),
    (Functions.ContentAddressableMemory, test_var, {'rate':RAND1, 'retrieval_prob':0.9, 'storage_prob':0.9}, [[
       0.5488135039273248, 0.7151893663724195, 0.6027633760716439, 0.5448831829968969, 0.4236547993389047, 0.6458941130666561, 0.4375872112626925, 0.8917730007820798, 0.9636627605010293, 0.3834415188257777], [
       0.7917250380826646, 0.5288949197529045, 0.5680445610939323, 0.925596638292661, 0.07103605819788694, 0.08712929970154071, 0.02021839744032572, 0.832619845547938, 0.7781567509498505, 0.8700121482468192 ]]),
    (Functions.ContentAddressableMemory, test_var, {'initializer':test_initializer, 'rate':RAND1}, [[
       0.5488135039273248, 0.7151893663724195, 0.6027633760716439, 0.5448831829968969, 0.4236547993389047, 0.6458941130666561, 0.4375872112626925, 0.8917730007820798, 0.9636627605010293, 0.3834415188257777], [
       0.7917250380826646, 0.5288949197529045, 0.5680445610939323, 0.925596638292661, 0.07103605819788694, 0.08712929970154071, 0.02021839744032572, 0.832619845547938, 0.7781567509498505, 0.8700121482468192 ]]),
# Disable noise tests for now as they trigger failure in ContentAddressableMemory lookup
#    (Functions.ContentAddressableMemory, test_var, {'rate':RAND1, 'noise':RAND2}, [[
#       0.79172504, 0.52889492, 0.56804456, 0.92559664, 0.07103606, 0.0871293 , 0.0202184 , 0.83261985, 0.77815675, 0.87001215 ],[
#       1.3230471933615413, 1.4894230558066361, 1.3769970655058605, 1.3191168724311135, 1.1978884887731214, 1.4201278025008728, 1.2118209006969092, 1.6660066902162964, 1.737896449935246, 1.1576752082599944
#]]),
#    (Functions.ContentAddressableMemory, test_var, {'rate':RAND1, 'noise':[RAND2], 'retrieval_prob':0.5},
#       [[ 0. for i in range(SIZE) ],[ 0. for i in range(SIZE) ]]),
#    (Functions.ContentAddressableMemory, test_var, {'rate':RAND1, 'noise':RAND2, 'storage_prob':0.5},
#       [[ 0. for i in range(SIZE) ],[ 0. for i in range(SIZE) ]]),
#    (Functions.ContentAddressableMemory, test_var, {'initializer':test_initializer, 'rate':RAND1, 'noise':RAND2}, [[
#       0.79172504, 0.52889492, 0.56804456, 0.92559664, 0.07103606, 0.0871293 , 0.0202184 , 0.83261985, 0.77815675, 0.87001215 ],[
#       1.3230471933615413, 1.4894230558066361, 1.3769970655058605, 1.3191168724311135, 1.1978884887731214, 1.4201278025008728, 1.2118209006969092, 1.6660066902162964, 1.737896449935246, 1.1576752082599944
#]]),
]

# use list, naming function produces ugly names
names = [
    "Buffer",
#    "Buffer Initializer",
    "ContentAddressableMemory",
    "ContentAddressableMemory Random Retrieval",
    "ContentAddressableMemory Random Storage",
    "ContentAddressableMemory Random Retrieval-Storage",
    "ContentAddressableMemory Initializer",
#    "ContentAddressableMemory Noise",
#    "ContentAddressableMemory Noise Random Retrieval",
#    "ContentAddressableMemory Noise Random Storage",
#    "ContentAddressableMemory Initializer Noise",
]

GROUP_PREFIX="IntegratorFunction "

@pytest.mark.function
@pytest.mark.memory_function
@pytest.mark.parametrize("func, variable, params, expected", test_data, ids=names)
@pytest.mark.benchmark
def test_basic(func, variable, params, expected, benchmark):
    f = func(default_variable=variable, **params)
    benchmark.group = GROUP_PREFIX + func.componentName
    f(variable)
    res = f(variable)
    assert np.allclose(res[0], expected[0])
    assert np.allclose(res[1], expected[1])
    benchmark(f, variable)


@pytest.mark.llvm
@pytest.mark.function
@pytest.mark.memory_function
@pytest.mark.parametrize("func, variable, params, expected", test_data, ids=names)
@pytest.mark.benchmark
def test_llvm(func, variable, params, expected, benchmark):
    if func is Functions.Buffer:
        pytest.skip("Not implemented")
    benchmark.group = GROUP_PREFIX + func.componentName
    f = func(default_variable=variable, **params)
    m = pnlvm.execution.FuncExecution(f)
    m.execute(variable)
    res = m.execute(variable)
    assert np.allclose(res[0], expected[0])
    assert np.allclose(res[1], expected[1])
    benchmark(m.execute, variable)


@pytest.mark.llvm
@pytest.mark.cuda
@pytest.mark.function
@pytest.mark.memory_function
@pytest.mark.parametrize("func, variable, params, expected", test_data, ids=names)
@pytest.mark.benchmark
def test_ptx_cuda(func, variable, params, expected, benchmark):
    if func is Functions.Buffer:
        pytest.skip("Not implemented")
    benchmark.group = GROUP_PREFIX + func.componentName
    f = func(default_variable=variable, **params)
    m = pnlvm.execution.FuncExecution(f)
    m.cuda_execute(variable)
    res = m.cuda_execute(variable)
    assert np.allclose(res, expected)
    benchmark(m.cuda_execute, variable)

# Test of ContentAddressableMemory without LLVM:
def test_ContentAddressableMemory_with_initializer_and_key_size_same_as_val_size():

    stimuli = {'A': [[1,2,3],[4,5,6]],
               'B': [[8,9,10],[11,12,13]],
               'C': [[1,2,3],[11,12,13]],
               'D': [[1,2,3],[21,22,23]],
               'E': [[9,8,4],[11,12,13]],
               'F': [[10,10,30],[40,50,60]],
               }

    em = EpisodicMemoryMechanism(
            content_size=3,
            assoc_size=3,
            function = ContentAddressableMemory(
                    seed=2,
                    initializer=np.array([stimuli['F'], stimuli['F']], dtype=object),
                    duplicate_keys=True,
                    equidistant_keys_select=RANDOM)
    )

    retrieved_keys=[]
    for key in sorted(stimuli.keys()):
        retrieved = [i for i in em.execute(stimuli[key])]
        retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
        retrieved_keys.append(retrieved_key)
    assert retrieved_keys == [['F'], ['A'], ['A'], ['C'], ['B'], ['F']]

    # Run again to test re-initialization and random retrieval
    em.function.reset(np.array([stimuli['A'], stimuli['F']]))
    retrieved_keys=[]
    for key in sorted(stimuli.keys()):
        retrieved = [i for i in em.execute(stimuli[key])]
        retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
        retrieved_keys.append(retrieved_key)
    assert retrieved_keys == [['A'], ['A'], ['A'], ['A'], ['B'], ['F']]

    stim = 'C'
    em.function.equidistant_keys_select = OLDEST
    retrieved = [i for i in em.function.get_memory(stimuli[stim][0])]
    retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
    assert retrieved_key == ['A']

    em.function.equidistant_keys_select = NEWEST
    retrieved = [i for i in em.function.get_memory(stimuli[stim][0])]
    retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
    assert retrieved_key == ['D']

    # Test that after allowing dups, warning is issued and memory with zeros is returned
    em.function.duplicate_keys = False
    stim = 'A'

    text = r'More than one item matched key \(\[1 2 3\]\) in memory for ContentAddressableMemory'
    with pytest.warns(UserWarning, match=text):
        retrieved = em.execute(stimuli[stim])

    retrieved_key = [k for k,v in stimuli.items() if v==list(retrieved)] or [None]
    assert retrieved_key == [None]
    assert retrieved[0] == [0, 0, 0]
    assert retrieved[1] == [0, 0, 0]

def test_ContentAddressableMemory_with_initializer_and_key_size_diff_from_val_size():

    stimuli = {'A': [[1,2,3],[4,5,6,7]],
               'B': [[8,9,10],[11,12,13,14]],
               'C': [[1,2,3],[11,12,13,14]],
               'D': [[1,2,3],[21,22,23,24]],
               'E': [[9,8,4],[11,12,13,14]],
               'F': [[10,10,30],[40,50,60,70]],
               }

    em = EpisodicMemoryMechanism(
            content_size=3,
            assoc_size=4,
            function = ContentAddressableMemory(
                    initializer=np.array([stimuli['F'], stimuli['F']], dtype=object),
                    duplicate_keys=True,
                    equidistant_keys_select=RANDOM)
    )

    retrieved_keys=[]
    for key in sorted(stimuli.keys()):
        print(key)
        retrieved = [i for i in em.execute(stimuli[key])]
        retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
        retrieved_keys.append(retrieved_key)
    assert retrieved_keys == [['F'], ['A'], ['A'], ['A'], ['B'], ['F']]

    stim = 'C'
    em.function.equidistant_keys_select = OLDEST
    retrieved = [i for i in em.function.get_memory(stimuli[stim][0])]
    retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
    assert retrieved_key == ['A']

    em.function.equidistant_keys_select = NEWEST
    retrieved = [i for i in em.function.get_memory(stimuli[stim][0])]
    retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
    assert retrieved_key == ['D']

    # Test that after allowing dups, warning is issued and memory with zeros is returned
    em.function.duplicate_keys = False
    stim = 'A'

    text = r'More than one item matched key \(\[1 2 3\]\) in memory for ContentAddressableMemory'
    with pytest.warns(UserWarning, match=text):
        retrieved = em.execute(stimuli[stim])

    retrieved_key = [k for k,v in stimuli.items() if v==list(retrieved)] or [None]
    assert retrieved_key == [None]
    assert retrieved[0] == [0, 0, 0]
    assert retrieved[1] == [0, 0, 0, 0]

def test_ContentAddressableMemory_without_initializer_and_key_size_same_as_val_size():

    stimuli = {'A': [[1,2,3],[4,5,6]],
               'B': [[8,9,10],[11,12,13]],
               'C': [[1,2,3],[11,12,13]],
               'D': [[1,2,3],[21,22,23]],
               'E': [[9,8,4],[11,12,13]],
               'F': [[10,10,30],[40,50,60]],
               }

    em = EpisodicMemoryMechanism(
            content_size=3,
            assoc_size=3,
            function = ContentAddressableMemory(
                    duplicate_keys=True,
                    equidistant_keys_select=RANDOM)
    )

    retrieved_keys=[]
    for key in sorted(stimuli.keys()):
        retrieved = [i for i in em.execute(stimuli[key])]
        retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
        retrieved_keys.append(retrieved_key)
    assert retrieved_keys == [[None], ['A'], ['A'], ['C'], ['B'], ['D']]

    stim = 'C'
    em.function.equidistant_keys_select = OLDEST
    retrieved = [i for i in em.function.get_memory(stimuli[stim][0])]
    retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
    assert retrieved_key == ['A']

    em.function.equidistant_keys_select = NEWEST
    retrieved = [i for i in em.function.get_memory(stimuli[stim][0])]
    retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
    assert retrieved_key == ['D']

    # Test that after allowing dups, warning is issued and memory with zeros is returned
    em.function.duplicate_keys = False
    stim = 'A'

    text = r'More than one item matched key \(\[1 2 3\]\) in memory for ContentAddressableMemory'
    with pytest.warns(UserWarning, match=text):
        retrieved = em.execute(stimuli[stim])

    retrieved_key = [k for k,v in stimuli.items() if v==list(retrieved)] or [None]
    assert retrieved_key == [None]
    assert retrieved[0] == [0, 0, 0]
    assert retrieved[1] == [0, 0, 0]

def test_ContentAddressableMemory_without_initializer_and_key_size_diff_from_val_size():

    stimuli = {'A': [[1,2,3],[4,5,6,7]],
               'B': [[8,9,10],[11,12,13,14]],
               'C': [[1,2,3],[11,12,13,14]],
               'D': [[1,2,3],[21,22,23,24]],
               'E': [[9,8,4],[11,12,13,14]],
               'F': [[10,10,30],[40,50,60,70]],
               }

    em = EpisodicMemoryMechanism(
            content_size=3,
            assoc_size=4,
            function = ContentAddressableMemory(
                    duplicate_keys=True,
                    equidistant_keys_select=RANDOM)
    )

    retrieved_keys=[]
    for key in sorted(stimuli.keys()):
        retrieved = [i for i in em.execute(stimuli[key])]
        retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
        retrieved_keys.append(retrieved_key)
    assert retrieved_keys == [[None], ['A'], ['A'], ['C'], ['B'], ['D']]

    stim = 'C'
    em.function.equidistant_keys_select = OLDEST
    retrieved = [i for i in em.function.get_memory(stimuli[stim][0])]
    retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
    assert retrieved_key == ['A']

    em.function.equidistant_keys_select = NEWEST
    retrieved = [i for i in em.function.get_memory(stimuli[stim][0])]
    retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
    assert retrieved_key == ['D']

    # Test that after allowing dups, warning is issued and memory with zeros is returned
    em.function.duplicate_keys = False
    stim = 'A'

    text = r'More than one item matched key \(\[1 2 3\]\) in memory for ContentAddressableMemory'
    with pytest.warns(UserWarning, match=text):
        retrieved = em.execute(stimuli[stim])

    retrieved_key = [k for k,v in stimuli.items() if v==list(retrieved)] or [None]
    assert retrieved_key == [None]
    assert retrieved[0] == [0, 0, 0]
    assert retrieved[1] == [0, 0, 0, 0]


def test_ContentAddressableMemory_without_assoc():

    stimuli = {'A': [[1,2,3]],
               'B': [[8,9,10]],
               'C': [[1,2,3]],
               'D': [[1,2,3]],
               'E': [[9,8,4]],
               'F': [[10,10,30]],
               }

    em = EpisodicMemoryMechanism(
            name='EPISODIC MEMORY MECH',
            content_size=3,
            function = ContentAddressableMemory(
                    # initializer=np.array([stimuli['F'], stimuli['F']], dtype=object),
                    duplicate_keys=True,
                    equidistant_keys_select=RANDOM,
                    retrieval_prob = 1.0
            )
    )

    for key in sorted(stimuli.keys()):
        print(f'\nCurrent memory: \n{em.memory}\n')
        retrieved = [i for i in em.execute(stimuli[key])]
        retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
        print(f'\nExecuted with stimulus {key}: {stimuli[key]};'
              f'\nRetrieved memory {retrieved_key[0]}: \n\t{retrieved}')

    retrieved_keys=[]
    for key in sorted(stimuli.keys()):
        retrieved = [i for i in em.execute(stimuli[key])]
        retrieved_key = [k for k,v in stimuli.items() if v == retrieved] or [None]
        retrieved_keys.append(retrieved_key)

    assert retrieved_keys == [['A', 'C', 'D'], ['B'], ['A', 'C', 'D'], ['A', 'C', 'D'], ['E'], ['F']]


def test_ContentAddressableMemory_with_duplicate_entry_in_initializer_warning():

    regexp = r'Attempt to initialize memory of ContentAddressableMemory with an entry \([[1 2 3]'
    with pytest.warns(UserWarning, match=regexp):
        em = EpisodicMemoryMechanism(
                name='EPISODIC MEMORY MECH',
                content_size=3,
                assoc_size=3,
                function = ContentAddressableMemory(
                        initializer=np.array([[[1,2,3], [4,5,6]],
                                              [[1,2,3], [7,8,9]]]),
                        duplicate_keys=False,
                        equidistant_keys_select=RANDOM,
                        retrieval_prob = 1.0
                )
        )
    assert np.allclose(em.memory, np.array([[[1, 2, 3], [4, 5, 6]]]))

def test_ContentAddressableMemory_add_and_delete_from_memory():

    em = ContentAddressableMemory(
            initializer=[[[1,2,3], [4,5,6]],
                         [[7,8,9], [10,11,12]]],
            duplicate_keys=True,
            equidistant_keys_select=RANDOM,
            retrieval_prob = 1.0,
            storage_prob = 1.0
    )
    em.add_to_memory([[[10,20,30],[40,50,60]],
                     [[11,21,31],[41,51,61]]])

    expected_memory = [[[ 1,  2,  3],[ 4,  5,  6]],
                       [[ 7,  8,  9],[10, 11, 12]],
                       [[10, 20, 30],[40, 50, 60]],
                       [[11, 21, 31],[41, 51, 61]]]
    assert np.allclose(em.memory, expected_memory)

    em.delete_from_memory([[[1,2,3],[4,5,6]]])
    expected_memory = [[[ 7,  8,  9],[10, 11, 12]],
                       [[10, 20, 30],[40, 50, 60]],
                       [[11, 21, 31],[41, 51, 61]]]
    assert np.allclose(em.memory, expected_memory)

    # Test adding and deleting a single memory
    em.add_to_memory([[1,2,3],[100,101,102]])
    expected_memory = [[[ 7,  8,  9],[10, 11, 12]],
                       [[10, 20, 30],[40, 50, 60]],
                       [[11, 21, 31],[41, 51, 61]],
                       [[ 1,  2,  3],[100,101,102]]]
    assert np.allclose(em.memory, expected_memory)

    em.delete_from_memory([[1,2,3],[100,101,102]])
    expected_memory = [[[ 7,  8,  9],[10, 11, 12]],
                       [[10, 20, 30],[40, 50, 60]],
                       [[11, 21, 31],[41, 51, 61]]]
    assert np.allclose(em.memory, expected_memory)

    # Test adding memory with different size value
    em.add_to_memory([[1,2,3],[100,101,102,103]])
    expected_memory = [[[ 7,  8,  9],[10, 11, 12]],
                       [[10, 20, 30],[40, 50, 60]],
                       [[11, 21, 31],[41, 51, 61]],
                       [[ 1,  2,  3],[100,101,102,103]]]
    for m,e in zip(em.memory,expected_memory):
        for i,j in zip(m,e):
            assert np.allclose(i,j)

    # Test adding memory with different size value as np.array
    em.add_to_memory(np.array([[1,2,3],[200,201,202,203]], dtype=object))
    expected_memory = [[[ 7,  8,  9],[10, 11, 12]],
                       [[10, 20, 30],[40, 50, 60]],
                       [[11, 21, 31],[41, 51, 61]],
                       [[ 1,  2,  3],[100,101,102,103]],
                       [[ 1,  2,  3],[200,201,202,203]]]
    for m,e in zip(em.memory,expected_memory):
        for i,j in zip(m,e):
            assert np.allclose(i,j)

    # Test error for illegal key:
    with pytest.raises(FunctionError) as error_text:
        em.add_to_memory(np.array([[1,2],[200,201,202,203]], dtype=object))
    assert "Length of 'key'" in str(error_text.value) and "must be same as others in the dict" in str(error_text.value)


def test_ContentAddressableMemory_overwrite_mode():

    em = ContentAddressableMemory(
            initializer=[[[1,2,3], [4,5,6]],
                         [[7,8,9], [10,11,12]]],
            duplicate_keys=True,
            equidistant_keys_select=RANDOM,
            retrieval_prob = 1.0,
            storage_prob = 1.0
    )

    em.duplicate_keys = OVERWRITE

    # Add new memory
    retreived = em.execute([[7,8,10], [100,110,120]])
    assert np.allclose(list(retreived), [[7,8,9],[10,11,12]])
    expected_memory = [[[ 1,  2,  3],[4, 5, 6]],
                       [[7,8,9], [10,11,12]],
                       [[7,8,10], [100,110,120]]]
    assert np.allclose(em.memory, expected_memory)

    # Overwrite old memory
    retreived = em.execute([[7,8,9], [100,110,120]])
    assert np.allclose(list(retreived), [[7,8,9],[10,11,12]])
    expected_memory = [[[ 1,  2,  3],[4, 5, 6]],
                       [[7,8,9], [100,110,120]],
                       [[7,8,10], [100,110,120]]]
    assert np.allclose(em.memory, expected_memory)

    # Allow entry of memory with duplicate key
    em.duplicate_keys = True
    retreived = em.execute([[7,8,9], [200,210,220]])
    assert np.allclose(list(retreived), [[7,8,9],[100,110,120]])
    expected_memory = [[[ 1,  2,  3],[4, 5, 6]],
                       [[7,8,9], [100,110,120]],
                       [[7,8,10], [100,110,120]],
                       [[7,8,9], [200,210,220]]]
    assert np.allclose(em.memory, expected_memory)

    # Attempt to overwrite with two matches should generate error
    em.duplicate_keys = OVERWRITE
    with pytest.raises(FunctionError) as error_text:
        em.execute([[7,8,9], [200,210,220]])
    assert ('Attempt to store item' in str(error_text.value)
            and 'with \'duplicate_keys\'=\'OVERWRITE\'' in str(error_text.value))


def test_ContentAddressableMemory_max_entries():

    em = ContentAddressableMemory(
            initializer=[[[1,2,3], [4,5,6]],
                         [[7,8,9], [10,11,12]],
                         [[1,2,3], [100,101,102]]],
            duplicate_keys=True,
            equidistant_keys_select=RANDOM,
            retrieval_prob = 1.0,
            storage_prob = 1.0,
            max_entries = 4
    )
    em.add_to_memory([[[10,20,30],[40,50,60]],
                    [[11,21,31],[41,51,61]],
                    [[12,22,32],[42,52,62]]])
    expected_memory = [[[1,2,3], [100,101,102]],
                       [[10,20,30],[40,50,60]],
                       [[11,21,31],[41,51,61]],
                       [[12,22,32],[42,52,62]]]
    assert np.allclose(em.memory, expected_memory)


@pytest.mark.parametrize(
    'param_name',
    [
        'distance_function',
        'selection_function',
    ]
)
def test_ContentAddressableMemory_unique_functions(param_name):
    a = ContentAddressableMemory()
    b = ContentAddressableMemory()

    assert (
        getattr(a.parameters, param_name).get()
        is not getattr(b.parameters, param_name).get()
    )

    assert (
        getattr(a.defaults, param_name)
        is not getattr(b.defaults, param_name)
    )



import functools
import numpy as np
import psyneulink as pnl
import psyneulink.core.components.functions.transferfunctions

Input_Layer = pnl.TransferMechanism(
    name='Input Layer',
    default_variable=np.zeros((2,)),
    function=psyneulink.core.components.functions.transferfunctions.Logistic
)

Hidden_Layer_1 = pnl.TransferMechanism(
    name='Hidden Layer_1',
    default_variable=np.zeros((5,)),
    function=psyneulink.core.components.functions.transferfunctions.Logistic()
)

Hidden_Layer_2 = pnl.TransferMechanism(
    name='Hidden Layer_2',
    default_variable=[0, 0, 0, 0],
    function=psyneulink.core.components.functions.transferfunctions.Logistic()
)

Output_Layer = pnl.TransferMechanism(
    name='Output Layer',
    default_variable=[0, 0, 0],
    function=psyneulink.core.components.functions.transferfunctions.Logistic
)

Gating_Mechanism = pnl.GatingMechanism(
    # default_gating_allocation=0.0,
    size=[1],
    gating_signals=[
        Hidden_Layer_1,
        Hidden_Layer_2,
        Output_Layer,
    ]
)

Input_Weights_matrix = (np.arange(2 * 5).reshape((2, 5)) + 1) / (2 * 5)
Middle_Weights_matrix = (np.arange(5 * 4).reshape((5, 4)) + 1) / (5 * 4)
Output_Weights_matrix = (np.arange(4 * 3).reshape((4, 3)) + 1) / (4 * 3)

# TEST PROCESS.LEARNING WITH:
# CREATION OF FREE STANDING PROJECTIONS THAT HAVE NO LEARNING (Input_Weights, Middle_Weights and Output_Weights)
# INLINE CREATION OF PROJECTIONS (Input_Weights, Middle_Weights and Output_Weights)
# NO EXPLICIT CREATION OF PROJECTIONS (Input_Weights, Middle_Weights and
# Output_Weights)

# This projection will be used by the process below by referencing it in the process' pathway;
#    note: sender and receiver args don't need to be specified
Input_Weights = pnl.MappingProjection(
    name='Input Weights',
    matrix=Input_Weights_matrix
)

# This projection will be used by the process below by assigning its sender and receiver args
#    to mechanismss in the pathway
Middle_Weights = pnl.MappingProjection(
    name='Middle Weights',
    sender=Hidden_Layer_1,
    receiver=Hidden_Layer_2,
    matrix={
        pnl.VALUE: Middle_Weights_matrix,
        pnl.FUNCTION: psyneulink.core.components.functions.statefulfunctions.integratorfunctions.AccumulatorIntegrator,
        pnl.FUNCTION_PARAMS: {
            pnl.INITIALIZER: Middle_Weights_matrix,
            pnl.RATE: Middle_Weights_matrix
        },
    }
)

Output_Weights = pnl.MappingProjection(
    name='Output Weights',
    sender=Hidden_Layer_2,
    receiver=Output_Layer,
    matrix=Output_Weights_matrix
)

z = pnl.Pathway(
    # default_variable=[0, 0],
    pathway=[
        Input_Layer,
        # The following reference to Input_Weights is needed to use it in the pathway
        # since it's sender and receiver args are not specified in its
        # declaration above
        Input_Weights,
        Hidden_Layer_1,
        # No projection specification is needed here since the sender arg for Middle_Weights
        #    is Hidden_Layer_1 and its receiver arg is Hidden_Layer_2
        # Middle_Weights,
        Hidden_Layer_2,
        # Output_Weights does not need to be listed for the same reason as Middle_Weights
        # If Middle_Weights and/or Output_Weights is not declared above, then the process
        #    will assign a default for missing projection
        # Output_Weights,
        Output_Layer
    ]
)

g = pnl.Pathway(
    pathway=[Gating_Mechanism]
)

stim_list = {
    Input_Layer: [[-1, 30]],
    Gating_Mechanism: [1.0]
}
target_list = {
    Output_Layer: [[0, 0, 1]]
}


def print_header(system):
    print("\n\n**** Time: ", system.scheduler.clock.simple_time)


def show_target():
    i = comp.input
    t = comp.target_input_ports[0].value
    print('\nOLD WEIGHTS: \n')
    print('- Input Weights: \n', Input_Weights.matrix)
    print('- Middle Weights: \n', Middle_Weights.matrix)
    print('- Output Weights: \n', Output_Weights.matrix)
    print('\nSTIMULI:\n\n- Input: {}\n- Target: {}\n'.format(i, t))
    print('ACTIVITY FROM OLD WEIGHTS: \n')
    print('- Middle 1: \n', Hidden_Layer_1.value)
    print('- Middle 2: \n', Hidden_Layer_2.value)
    print('- Output:\n', Output_Layer.value)


comp = pnl.Composition(
    pathways=[z, g],
    targets=[0, 0, 1],
    learning_rate=1.0
)

comp.reportOutputPref = True
# mySystem.show_graph(show_learning=True)

results = comp.learn(
    num_trials=10,
    inputs=stim_list,
    targets=target_list,
    clamp_input=pnl.SOFT_CLAMP,
    learning_rate=1.0,
    prefs={
        pnl.VERBOSE_PREF: False,
        pnl.REPORT_OUTPUT_PREF: True
    },
    call_before_trial=functools.partial(print_header, comp),
    call_after_trial=show_target,
)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************  ModulatorySignal *******************************************************

"""

Contents
--------
  * `ModulatorySignal_Overview`
  * `ModulatorySignal_Creation`
  * `ModulatorySignal_Structure`
      - `ModulatorySignal_Projections`
      - `ModulatorySignal_Modulation`
          • `ModulatorySignal_Uses`
          • `ModulatorySignal_Types`
          • `ModulatorySignal_Implementation`
  * `ModulatorySignal_Execution`
  COMMENT:
  * `ModulatorySignal_Examples`
  COMMENT
  * `ModulatorySignal_Class_Reference`


.. _ModulatorySignal_Overview:

Overview
--------

A ModulatorySignal is a subclas of `OutputPort` that belongs to a `ModulatoryMechanism <ModulatoryMechanism>`, and is
used to `modulate <ModulatorySignal_Modulation>` the `value <Port_Base.value>` of one or more `Ports <Port>` by way of
one or more `ModulatoryProjctions <ModulatoryProjection>`. A ModulatorySignal modulates the value of a Port by modifying
a  parameter of that Port's `function <Port_Base.function>`.  There are three types of ModulatorySignals, each of which
is  associated wth a particular type of `ModulatoryMechanism <ModulatoryMechanism>` and `ModulatoryProjection
<ModulatoryProjection>`, and modifies the value of different types of Ports, as summarized `below:

* `ControlSignal`
    takes the `allocation <ControlSignal.allocation>` assigned to it by the `function <ControlMechanism.function>`
    of the `ControlMechanism <ControlMechanism>` to which it belongs, and uses it to modulate the parameter of a
    `Mechanism <Mechanism>` or its `function <Mechanism_Base.function>` (and thereby the `value
    <Mechanism_Base.value>` of that Mechanism), or a parameter of the `function <Port_Base.function>` one of the
    Mechanism's `InputPorts <InputPort>` or `OutputPorts <OutputPort>` (and thereby the `value <Port_Base.value>`
    of the corresponding Port).
..
* `GatingSignal` takes the `allocation <GatingSignal.allocation>` assigned to it by the `function
    <GatingMechanism.function>` of the `GatingMechanism` to which it belongs, and uses it to modulate the parameter
    of the `function <Port_Base.function>` of an `InputPort` or `OutputPort` (and hence that Port's `value
    <Port_Base.value>`).  A GatingMechanism and GatingSignal can be thought of as implementing a form of control
    specialized for gating the input to and/or output of a Mechanism.
..
* `LearningSignal`
    takes the `learning_signal <LearningMechanism>` calculated by the `function <LearningMechanism.function>` of the
    `LearningMechanism` to which it belongs, and uses it to modulate the `matrix <MappingProjection.matrix>` parameter
    of a `MappingProjection`.

These are shown in a `figure <ModulatorySignal_Anatomy_Figure>` below, and are descdribed in greater detail in
in the sections under `ModulatorySignal_Structure`.

See `ModulatoryMechanism <ModulatoryMechanism_Naming>` for conventions used for the names of Modulatory components.

.. _ModulatorySignal_Creation:

Creating a ModulatorySignal
---------------------------

ModulatorySignal is a base class, and cannot be instantiated directly.  However, the three types of ModulatorySignals
listed above can be created directly, by calling the constructor for the desired type.  More commonly, however,
ModulatorySignals are created automatically by the `ModulatoryMechanism <ModulatoryMechanism>` to which they belong, or
by specifying them in the constructor for a `ModulatoryMechanism <ModulatoryMechanism>` (the details of which are
described in the documentation for each type of ModulatorySignal).  If a ModulatorySignal is constructed explicitly,
the type of modulation it uses is specifed in the **modulation** argument of its constructor, using a 2-item tuple
that contains the Port to be modulated as the first item, and either the name of the parameter of the Port's `function
<Port_Base.function>` to be modulated, or a keyword specifying the type of modulation, as the second item (see
`ModulatorySignal_Types` for additional details).

.. _ModulatorySignal_Structure:

Structure
---------

A ModulatorySignal is always assigned to a `ModulatoryMechanism <ModulatoryMechanism>`, and must be assigned to an
ModulatoryMechanism of the appropriate type (see `Types of ModulatoryMechanism <ModulatoryMechanism_Types>`).  The
ModulatorySignal receives an `allocation <ModulatorySignal.allocation>` from the ModulatoryMechanism to which it is
assigned, that it uses as the `variable <Function_Base.variable>` for its `function <Mechanism_Base.function>`, the
result of which is the modulatory `value <ModulatorySignal.value>` of the ModulatorySignal.  A ModulatorySignal is
associated with one or more `ModulatoryProjections <ModulatoryProjection>` of the corresponding type, that receive
the ModulatorySignal's `value <ModulatorySignal.value>`, and use this to modulate the Port(s) to which they project.
All of the ModulatoryProjections from a given ModulatorySignal are assigned the same modulatory `value
<ModulatorySignal.value>` (see `ModulatorySignal_Projections` below) and use the same `type of modulation
<ModulatorySignal_Types>` specified by the the ModulatorySignal's `modulation <ModulatorySignal.modulation>` attribute.
The ModulatoryProjections received by a `Port <Port>` are listed in the Port's `mod_afferents <Port_Base.mod_afferents>`
attribute.

The section on `Modulation` below provides a comparison of ModulatorySignal `subclasses and their uses
<ModulatorySignal_Uses>` (summarized in an accompanying `table <ModulatorySignal_Table>` and `figure
<ModulatorySignal_Anatomy_Figure>`), as well as a description of the different `types of modulation
<ModulatorySignal_Types>` and a more detailed description of `how modulation is implemented
<ModulatorySignal_Implementation>`.

.. _ModulatorySignal_Projections:

*ModulatoryProjections*
~~~~~~~~~~~~~~~~~~~~~~~

A ModulatorySignal can be assigned one or more `ModulatoryProjections <ModulatoryProjection>`,
using either the **projections** argument of its constructor, or in an entry of a dictionary assigned to the
**params** argument with the key *PROJECTIONS*.  These are assigned to its `efferents  <ModulatorySignal.efferents>`
attribute.  See `Port Projections <Port_Projections>` for additional details concerning the specification of
Projections when creating a Port.

Although a ModulatorySignal can be assigned more than one `ModulatoryProjection <ModulatoryProjection>`,
all of those Projections receive and convey the same modulatory `value <ModulatorySignal.value>` from the
ModulatorySignal, and use the same form of `modulation <ModulatorySignal_Modulation>`.  This is a common use for some
ModulatorySignals (e.g., the use of a single `GatingSignal` to gate multiple `InputPort(s) <InputPort>` or
`OutputPort(s) <OutputPort>`), but requires more specialized circumstances for others (e.g., the use of a single
`LearningSignal` for more than one `MappingProjection`, or a single `ControlSignal` for the parameters of more than
one Mechanism or function).

.. _ModulatorySignal_Modulation:

*Modulation*
~~~~~~~~~~~~

A ModulatorySignal modulates the value of a `Port <Port>` either by modifying a parameter of the Port's `function
<Port_Base.function>` (which determines the Port's `value <Port_Base.value>`), or by  assigning a value to the Port
directly.  The `type of modulation <ModulatorySignal_Types>` is determined by the ModulatorySignal's
`modulation <ModulatorySignal.modulation>` attribute, which can be specified in the **modulation** argument of its
ModulatorySignal's constructor, or in a *MODULATION* entry of a `Port specification dictionary
<Port_Specification>` used to create the ModulatorySignal (see `Type of Modualtion <ModulatorySignal_Types>` and
`figure <ModulatorySignal_Detail_Figure>` below for details). If the type of `modulation <ModulatorySignal.modulation>`
is not specified when a ModulatorySignal is created, it is assigned the value of the `modulation
<ModulatoryMechanism_Base.modulation>` attribute for the `ModulatoryMechanism <ModulatoryMechanism>` to which it
belongs.

.. _ModulatorySignal_Uses:

Uses of Modulation
^^^^^^^^^^^^^^^^^^

There are three broad types of modulation that serve different purposes, and differ according to the
ModulatorySignals used and the type of Port modulated;  these are modulation of a:

  * `Mechanism's  <Mechanism>` `function <Mechanism_Base.function>`
        a `ControlSignal` must be used; this modulates the `value <ParameterPort.value>` of `ParameterPort` for a
        parameter of the Mechanism's `function <Mechanism_Base.function>` which, in turn, determines how it computes
        the Mechanism's `value <Mechanism_Base.value>`;

  * `Mechanism's <Mechanism>` input or output
        a `GatingSignal` is specialized for this purpose, though a `ControlSignal` can also be used.  These modulate
        either the value <InputPort.value>` an `InputPort` of the Mechanism, that determines the Mechanism's `variable
        <Mechanism_Base.variable>` used as the input to its `function <Mechanism_Base.function>`; or the `value
        <OutputPort.value>` of an `OutputPort` of the Mechanism, that determines how the `value <Mechanism_Base.value>`
        of the Mechanism (i.e., the result of its `function <Mechanism_Base.function>`) is used to generate the output
        from that Mechanism.

  * `MappingProjection`
        a `LearningSignal` must be used; this modulates the `ParameterPort` for the `matrix <MappingProjection.matrix>`
        parameter of a MappingProjection's `function <Projection_Base.function>` which, in turn, determines how it
        computes the MappingProjection's `value <Projection_Base.value>`.

The following table summarizes the three uses of modulation, the ModulatorySignals for each, and the Ports they
modulate.  The mechanics of modulation are described in greater detail in `ModulatorySignal_Implementation`,
and shown in the  `figure below <ModulatorySignal_Anatomy_Figure>`.

.. _ModulatorySignal_Table:

**ModulatorySignals and Ports they Modulate**  (colors listed are those used in the `figure <ModulatorySignal_Anatomy_Figure>` below)

.. table::
  :align: left

+------------------------------------+------------------------+------------------------------+---------------------------------------+----------------------------+
|                                    |                        |Default type of `modulation   |                                       |Default Function (mod param)|
|             Purpose                |  ModulatorySignal      |<ModulatorySignal.modulation>`|           Recipient Port              |for Recipient Port          |
+====================================+========================+==============================+=======================================+============================+
| Modulate the parameter of a        |                        |                              | Mechanism `ParameterPort` (by default)|                            |
| Mechanism's `function              | `ControlSignal` (blue) |     *MULTIPLICATIVE*         | but can also be an                    |     `Linear` (`slope`)     |
| <Mechanism_Base.function>`         |                        |                              | `InputPort` or `OutputPort`           |                            |
+------------------------------------+------------------------+------------------------------+---------------------------------------+----------------------------+
| Modulate the input or output of    |                        |                              |                                       |                            |
| a Mechanism's `function            | `GatingSignal` (brown) |     *MULTIPLICATIVE*         |  Mechanism `InputPort`/`OutputPort`   |    `Linear` (`slope`)      |
| <Mechanism_Base.function>`         |                        |                              |                                       |                            |
+------------------------------------+------------------------+------------------------------+---------------------------------------+----------------------------+
| Modulate a MappingProjection's     |                        |                              |                                       |   `AccumulatorIntegrator`  |
| `matrix <MappingProjection.matrix>`|`LearningSignal` (green)|        *ADDITIVE*            |  MappingProjection `ParameterPort`    |   (`increment`)            |
| parameter                          |                        |                              |                                       |                            |
+------------------------------------+------------------------+------------------------------+---------------------------------------+----------------------------+

It is important to emphasize that, although the purpose of a ModulatorySignal is to modify the functioning of a
`Mechanism <Mechanism>` or a `MappingProjection`, it does this indirectly by modifying a Port that determines the input
or output of a Mechanism, or the parameters of a Mechanism or Projection's `function`, rather than directly modifying
the function of the Mechanism or Projection itself.  This is shown in the following figure, and described in greater
detail under `ModulatorySignal_Implementation`.

.. _ModulatorySignal_Anatomy_Figure:

**Anatomy of Modulation**

.. figure:: _static/Modulation_Anatomy_fig.svg
   :alt: Modulation
   :scale: 150 %

   **Three types of Modulatory Components and the Ports they modulate**. The default `type of modulation
   <ModulatorySignal_Types>` for each type of ModulatorySignal, and the default Function and modulated parameter of
   its recipient Port are listed in the `table <ModulatorySignal_Table>` above. Note that the `ControlMechanism`
   and `ControlSignal <ControlSignal>` are shown in the figure modulating the `ParameterPort` of a Mechanism;
   however, like Gating components, they can also be used to modulate `InputPorts <InputPort>` and `OutputPorts
   <OutputPort>`. The `figure <ModulatorySignal_Detail_Figure>` below shows a detailed view of how ModulatorySignals
   modulate the parameters of a Port's `function <Port_Base.function>`.


.. _ModulatorySignal_Types:

Types of Modulation
^^^^^^^^^^^^^^^^^^^

The `modulation <ModulatorySignal.modulation>` attribute of a ModulatorySignal determines the way in which it
modulates the `value <Port_Base.value>` of a `Port <Port>`, by specifying which parameter of the Port's `function
<Port_Base.function>` that it modifies (see `figure <ModulatorySignal_Detail_Figure>` below).  This is specified
in a tuple containing the Port and the name of the parameter to be modified (see `example
<ControlSignal_Example_Modulate_Costs>`). Alternatively, a keyword can be used in place of the parameter's name.
For some `Functions <Function>`, keywords can be used to specify function-specific forms of modulation (e.g., see
`TransferWithCosts Function <TransferWithCosts_Modulation_of_Cost_Params>` for an example).  In addition, there are
four keywords that can be used to specify generic forms of modulation supported by most `Functions <Function>`:

  * *MULTPLICATIVE_PARAM* - assign the `value <ModulatorySignal.value>` of the ModulatorySignal to the parameter of
    the Port's `function <Port_Base.function>` specified as its `multiplicative_param <Function_Modulatory_Params>`.
    For example, if the Port's `function <Port_Base.function>` is `Linear` (the default for most Ports), then
    the ModulatorySignal's `value <ModulatorySignal.value>` is assigned to the function's `slope <Linear.slope>`
    parameter (it's multiplicative_param), thus multiplying the Port's `variable <Port_Base.variable>` by that
    amount each time the Port is executed, and assigning the result as the Port's `value <Port_Base.value>`.

  * *ADDITIVE_PARAM* - assign the `value <ModulatorySignal.value>` of the ModulatorySignal to the parameter of the
    Port's `function <Port_Base.function>` specified as its `additive_param <Function_Modulatory_Params>`. For
    example, if the Port's `function <Port_Base.function>` is `Linear` (the default for most Ports), then the
    ModulatorySignal's `value <ModulatorySignal.value>` is assigned to the function's `intercept <Linear.intercept>`
    parameter (it's additive_param), thus adding that value to the Port's `variable <Port_Base.variable>` each
    time the Port is executed, and assigning the result as the Port's `value <Port_Base.value>`.

  * *OVERRIDE* - assign the `value <ModulatorySignal.value>` of the ModulatorySignal directly to the Port's
    `value <Port_Base.value>`; in effect, this bypasses the Port's `function <Port_Base.function>`. Note that
    this can be specified for **only one** `ModulatorySignal` that modulates a given Port (see `below
    <ModulatorySignal_Multiple>` for additional details).

  * *DISABLE* - suppresses the modulatory effect of the ModulatorySignal;  the Port's `function <Port_Base.function>`
    will operate as if it did not receive a `ModulatoryProjection <ModulatoryProjection>` from that ModulatorySignal.

   .. note::

      the *MULTPLICATIVE_PARAM* and *ADDITIVE_PARAM* keywords can be used only with `Functions <Function>` that
      specify a `multiplicative_param and/or additive_param <Function_Modulatory_Params>`, respectively.

COMMENT:
FOR DEVELOPERS:  the MULTPLICATIVE_PARAM and ADDITIVE_PARAM options above are keywords for aliases to the relevant
parameters of a given Function, declared in its Parameters subclass declaration of the Function's declaration.
COMMENT

The default type of modulation for `ControlSignals <ControlSignal>` and `GatingSignals <GatingSignal>` is
*MULTIPLICATIVE*.  The default for `LearningSignals <LearningSignal>` is *ADDITIVE* (which additively modifies the
`value <LearningSignal.value>` of the LearningSignal (i.e., the weight changes computed by the `LearningMechanism`)
to the Port's `variable <Port_Base.variable>` (i.e., the current weight `matrix <MappingProjection.matrix>` for
the `MappingProjection` being learned).


.. _ModulatorySignal_Implementation:

Implementation of Modulation
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Although the `purpose of a ModulatorySignal <ModulatorySignal_Uses>` is to modify the operation of a `Mechanism
<Mechanism>` or `MappingProjection`, it does not do this *directly*; rather, it does it by way of a `Port` that
modulates the input, function parameter, or output of the Mechanism or MappingProjeciton to be regulated. More
specifically, a ModulatorySignal modulates the `function <Port_Base.function>` of a *Port* responsible for
generating those values, by modifying a parameter of that Port's `function <Port_Base.function>`.  This is shown
in the figure below:

.. _ModulatorySignal_Detail_Figure:

**Detailed View of Modulation**

.. figure:: _static/Modulation_Detail_fig.svg
   :alt: Modulation_Detail
   :scale: 150 %

   A ModulatorySignal modulates the `value <Port_Base.value>` of a Port either by modifying a parameter of the
   Port's `function <Port_Base.function>`, or assigining the `value <Port_Base.value>` of the Port directly.
   This is determined by the ModulatorySignal's `modulation <ModulatorySignal.modulation>` attribute.  That can be
   assigned either the name of a parameter of the Port's `function <Port_Base.function>`, or a keyword that
   specifies a standard form of modulation.  The keywords *MULTIPLICATIVE* and *ADDITIVE* specify that the `value
   <ModulatorySignal.value>` of the ModulatorySignal be assigned to the `multiplicative_param or `additive_param
   <Function_Modulatory_Params>` of the Port's function, respectively;  *OVERRIDE* specifies that the
   ModulatorySignal's `value <ModulatorySignal.value>` be assigned directly as the Port's `value
   <Port_Base.value>`, in effect bypassing the Port's `function <Port_Base.function>` (see
   `ModulatorySignal_Types` for additional details).

Though this implementaton of modulation is indirect, it provides a standard for all forms of modulation, as well as
considerable flexibility in the modulatory regulation of Components within a `Composition` (see
`ModulatorySignal_Types` below).

The types of Ports modulated by each type of ModulatorySignal are summarized in `ModulatorySignal_Uses`,
and the accompanying `table <ModulatorySignal_Table>` and `figure <ModulatorySignal_Anatomy_Figure>`.

.. _ModulatorySignal_Multiple:

Any `modulable <Parameter.modulable>` parameter of a Port's `function <Port_Base.function>` can be modulated,
and different parameters of the same `function <Port_Base.function>` of a Port can be modulated by different
ModulatorySignals. The same parameter can also be modulated by more than on ModulatorySignal. If more than one
ModulatorySignal modulates the same parameter of a Port's `function <Port_Base.function>`, then that parameter's
`modulation_combine_function <Parameter.modulation_combine_function>` attribute determines how the `value
<ModulatorySignal.value>`\\s of the different ModulatorySignals are combined.  By default, the product of their
vaues is used.  However, if *OVERRIDE* is specfied as the type of `modulation <ModulatorySignal.modulation>` for one
of them, then that ModulatorySignal's  `value <ModulatorySignal.value>` is assigned directly as the Port's `value
<Port_Base.value>`, and the others are all ignored.  Only one ModulatorySignal specified as *OVERRIDE* can modulate
a given parameter;  if there is more than, then an error is generated.

.. _ModulatorySignal_Execution:

Execution
---------

ModulatorySignals cannot be executed directly.  This done when the `ModulatoryMechanism <ModulatoryMechanism>` to
which they belong is executed. When a ModulatorySignal is executed, it calculates its `value <ModulatorySignal.value>`,
which is then assigned as the `variable <Projection_Base.variable>` of the `ModulatoryProjections
<ModulatoryProjection>` listed in its `efferents <ModulatorySignal.efferents>` attribute.
When those Projections execute, they convey the ModulatorySignal's `value <ModulatorySignal.value>` to the `function
<Port_Base.function>` of the `Port <Port>` to which they project.  The Port's `function <Port_Base.function>`
then uses that value in determining value of the parameter designated by the `modulation <ModulatorySignal.modulation>`
attribute of the ModulatorySignal when the Port's `value <Port_Base.value>` is updated.

COMMENT:

# FIX: 9/3/19 -- REWORK AND ADD EXAMPLE HERE

For example, consider a `ControlSignal` that modulates the `bias <Logistic.bias>` parameter of a `Logistic` Function
used by a `TransferMechanism`, and assume that the `ParameterPort` for the bias parameter (to which the ControlSignal
projects) uses a `Linear` function (the default for a ParameterPort) to set the `value <ParameterPort.value>` of
that parameter. If the `modulation  <ModulatorySignal.modulation>` attribute of the `ControlSignal` is *MULTIPLICATIVE*
then, when the TransferMechanism's `Logistic` `function <Mechanism_Base.function>` is executed, the `function
<ParameterPort.function>` of the ParameterPort that sets the value of the `Logistic` Function's `bias <Logistic.bias>`
parameter is executed;  that is a `Linear` Function, that uses the ControlSignal's `value <ControlSignal.value>` as
its `slope <Linear.slope>` parameter.  Thus, the effect is that the ControlSignal's `value <ControlSignal.value>` is
multiplied by the base value of the `bias <Logistic.bias>` parameter, before that is used by the TransferMechanism's
`Logistic` Function.  Thus, the `value <ControlSignal.value>` of the ControlSignal modulates the `bias
<Logistic.bias>` parameter of the `Logistic` Function when the TransferMechanism's `function
<Mechanism_Base.function>` is executed (see `Port Execution <State_Execution>` for additional details).

COMMENT

# FIX 5/8/20 -- REWORK TO BE ALIGNED WITH ModulatoryMechanism

.. note::
   The change in the value of a `Port <Port>` in response to a ModulatorySignal does not occur until the Mechanism to
   which the port belongs is next executed; see `Lazy Evaluation <Component_Lazy_Updating>` for an explanation of
   "lazy" updating).

COMMENT:

.. _ModulatorySignal_Examples:

Examples
--------

MOVE THESE TO SPECIFIC SUBCLASSES? AND REFERENCE THOSE HERE
FIX: EXAMPLE OF FULL SPECIFIATION (BY PORT AND PORT'S FUCNTION'S PARAMETER NAME)
The following example uses a parameter's name to specify
    >>> my_mech = ProcessingMechanism(function=Logistic)
    >>> ctl_mech = ControlMechanism(monitor_for_control=my_mech,
    ...                             control_signals=ControlSignal(modulates=my_mech.parameter_ports[GAIN],
    ...                                                           modulation=SLOPE))

FIX: EXAMPLE OF SPECIFIATION OF CONTROLSIGNAL WITH MECHANISM AND PORT'S PARAMETER NAME

FIX: EXAMPLE OF SPECIFIATION BY CONTROLSIGNAL WITH MECHANISM AND MECHANISM'S PARAMETER NAME

MENTION PORT-SPECIFIC CONVENIENCE METHODS

FIX: EXAMPLE OF CONTROL SIGNAL MODULATION OF INPUTPORT


.  For
example, the `TransferWithCosts` Function defines keywords for `modulating the parameters of its cost functions
<TransferWithCosts_Modulation_of_Cost_Params>`.
A ControlMechanism can even modulate the parameters of another
ControlMechanism, or its ControlSignals.  For example, in the following, ``ctl_mech_A`` modulates the `intensity_cost
<ControlSignal.intensity_cost>` parameter of ``ctl_mech``\\'s ControlSignal::

    >>> my_mech = ProcessingMechanism()
    >>> ctl_mech_A = ControlMechanism(monitor_for_control=my_mech,
    ...                               control_signals=ControlSignal(modulates=(SLOPE,my_mech),
    >>>                                                              cost_options = CostFunctions.INTENSITY))
    >>> ctl_mech_B = ControlMechanism(monitor_for_control=my_mech,
    ...                               control_signals=ControlSignal(modulates=ctl_mech_A.control_signals[0],
    ...                                                             modulation=INTENSITY_COST_FCT_MULTIPLICATIVE_PARAM))

FIX: EXAMPLE HERE
     THEN, AFTER EXAMPLE: MODULATORYSIGNAL'S VALUE IS ASSIGNED TO THE SPECIFIED PARAMETER.
COMMENT

.. _ModulatorySignal_Class_Reference:

Class Reference
---------------
"""

from psyneulink.core.components.component import component_keywords
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.keywords import \
    ADDITIVE_PARAM, DISABLE, MAYBE, MECHANISM, MODULATION, MODULATORY_SIGNAL, MULTIPLICATIVE_PARAM, \
    OVERRIDE, PROJECTIONS, VARIABLE

from psyneulink.core.globals.defaults import defaultModulatoryAllocation
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel

__all__ = [
    'modulatory_signal_keywords', 'ModulatorySignal', 'ModulatorySignalError',
]


def _is_modulatory_spec(spec, include_matrix_spec=True):
    from psyneulink.core.components.mechanisms.modulatory.learning.learningmechanism import _is_learning_spec
    from psyneulink.core.components.mechanisms.modulatory.control.controlmechanism import _is_control_spec
    from psyneulink.core.components.mechanisms.modulatory.control.gating.gatingmechanism import _is_gating_spec

    if (_is_learning_spec(spec, include_matrix_spec=include_matrix_spec)
        or _is_control_spec(spec)
        or _is_gating_spec(spec)
        ):
        return True
    else:
        return False

modulatory_signal_keywords = {MECHANISM, MODULATION}
modulatory_signal_keywords.update(component_keywords)
modulation_type_keywords = [MULTIPLICATIVE_PARAM, ADDITIVE_PARAM, OVERRIDE, DISABLE]


class ModulatorySignalError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


class ModulatorySignal(OutputPort):
    """Subclass of `OutputPort` used by a `ModulatoryMechanism <ModulatoryMechanism>` to modulate the value
    of one more `Ports <Port>`.  See `OutputPort <OutputPort_Class_Reference>` and subclasses for additional
    arguments and attributes.

    .. note::
       ModulatorySignal is an abstract class and should *never* be instantiated by a call to its constructor.
       It should be instantiated using the constructor for a `subclass <ModulatorySignal_Subtypes>`.

    COMMENT:
    PortRegistry
    -------------
        All OutputPorts are registered in PortRegistry, which maintains an entry for the subclass,
        a count for all instances of it, and a dictionary of those instances
    COMMENT


    Arguments
    ---------

    default_allocation : scalar : defaultModulatoryAllocation
        specifies the default template and value used for `variable <ModulatorySignal.variable>`.

    modulation : str : default MULTIPLICATIVE
        specifies the type of modulation the ModulatorySignal uses to determine the value of the Port(s) it modulates;
        must be either a keyword defined by the `Function` of the parameter to be modulated, or one of the following
        generic keywords -- *MULTIPLICATIVE*, *ADDITIVE*, *OVERRIDE* or *DISABLE* (see `ModulatorySignal_Types` for
        additional details).

    Attributes
    ----------

    variable : scalar, list or np.ndarray
        same as `allocation <ModulatorySignal.allocation>`.

    allocation : float
        value assigned by the ModulatorySignal's `owner <ModulatorySignal.owner>`, and used as the `variable
        <ModulatorySignal.variable>` of its `function <ModulatorySignal.function>` to determine the ModulatorySignal's
        `ModulatorySignal.value`.
    COMMENT:
    FOR DEVELOPERS:  Implemented as an alias of the ModulatorySignal's variable Parameter
    COMMENT

    function : TransferFunction
        used to transform the ModulatorySignal's `allocation <ModulatorySignal.allocation>` into its `value
        <ModulatorySignal.value>`;  default is the `Identity` Function that simply assigns `allocation
        <ModulatorySignal.allocation>` to `value <ModulatorySignal.value>`.

    value : number, list or np.ndarray
        result of `function <ModulatorySignal.function>`, used to determine the `value <Port_Base.value>` of the Port(s)
        being modulated.

    modulation : str
        determines how the `value <ModulatorySignal.value>` of the ModulatorySignal is used to modulate the value of
        the port(s) being modulated (see `ModulatorySignal_Types` for additional details).

    efferents : [List[GatingProjection]]
        a list of the `ModulatoryProjections <ModulatoryProjection>` assigned to the ModulatorySignal.

    name : str
        the name of the ModulatorySignal. If the ModulatorySignal's `initialization has been deferred
        <Port_Deferred_Initialization>`, it is assigned a temporary name (indicating its deferred initialization
        status) until initialization is completed, at which time it is assigned its designated name.  If that is the
        name of an existing ModulatorySignal, it is appended with an indexed suffix, incremented for each Port with
        the same base name (see `Registry_Naming`). If the name is not  specified in the **name** argument of its
        constructor, a default name is assigned as follows; if the ModulatorySignal has:

        * no projections (which are used to name it) -- the name of its class is used, with an index that is
          incremented for each ModulatorySignal with a default named assigned to its `owner <ModulatorySignal.owner>`;

        * one `ModulatoryProjection <ModulatoryProjction>` -- the following template is used:
          "<target Mechanism name> <target Port name> <ModulatorySignal type name>"
          (for example, ``'Decision[drift_rate] ControlSignal'``, or ``'Input Layer[InputPort-0] GatingSignal'``);

        * multiple ModulatoryProjections, all to Ports of the same Mechanism -- the following template is used:
          "<target Mechanism name> (<target Port name>,...) <ModulatorySignal type name>"
          (for example, ``Decision (drift_rate, threshold) ControlSignal``, or
          ``'Input Layer[InputPort-0, InputPort-1] GatingSignal'``);

        * multiple ModulatoryProjections to Ports of different Mechanisms -- the following template is used:
          "<owner Mechanism's name> divergent <ModulatorySignal type name>"
          (for example, ``'ControlMechanism divergent ControlSignal'`` or ``'GatingMechanism divergent GatingSignal'``).

        .. note::
            Unlike other PsyNeuLink components, Port names are "scoped" within a Mechanism, meaning that Ports with
            the same name are permitted in different Mechanisms.  However, they are *not* permitted in the same
            Mechanism: Ports within a Mechanism with the same base name are appended an index in the order of their
            creation.

    """

    componentType = MODULATORY_SIGNAL
    # paramsType = OUTPUT_PORT_PARAMS

    class Parameters(OutputPort.Parameters):
        """
            Attributes
            ----------

                modulation
                    see `modulation <ModulatorySignal_Modulation>`

                    :default value: None
                    :type:
        """
        modulation = None

    portAttributes = OutputPort.portAttributes | {MODULATION}

    classPreferenceLevel = PreferenceLevel.TYPE
    # Any preferences specified below will override those specified in TYPE_DEFAULT_PREFERENCES
    # Note: only need to specify setting;  level will be assigned to TYPE automatically
    # classPreferences = {
    #     PREFERENCE_SET_NAME: 'OutputPortCustomClassPreferences',
    #     PREFERENCE_KEYWORD<pref>: <setting>...}

    def __init__(self,
                 owner=None,
                 size=None,
                 reference_value=None,
                 default_allocation=defaultModulatoryAllocation,
                 function=None,
                 modulates=None,
                 modulation=None,
                 index=None,
                 assign=None,
                 params=None,
                 name=None,
                 prefs=None,
                 **kwargs):

        if kwargs:
            if VARIABLE in kwargs:
                default_allocation = kwargs.pop(VARIABLE, default_allocation)
            if PROJECTIONS in kwargs:
                modulates = kwargs.pop(PROJECTIONS, modulates)

        if params is not None:
            if PROJECTIONS in params:
                modulates = params.pop(PROJECTIONS, modulates)

        # Deferred initialization
        # if self.initialization_status & (ContextFlags.DEFERRED_INIT | ContextFlags.INITIALIZING):
        if self.initialization_status & ContextFlags.DEFERRED_INIT:
            # If init was deferred, it may have been because owner was not yet known (see OutputPort.__init__),
            #   and so modulation hasn't had a chance to be assigned to the owner's value
            #   (i.e., if it was not specified in the constructor), so do it now;
            #   however modulation has already been assigned to params, so need to assign it there
            modulation = modulation or owner.modulation

        if modulates is not None and not isinstance(modulates, list):
            modulates = [modulates]

        super().__init__(owner=owner,
                         reference_value=reference_value,
                         variable=default_allocation,
                         size=size,
                         projections=modulates,
                         index=index,
                         assign=assign,
                         function=function,
                         modulation=modulation,
                         params=params,
                         name=name,
                         prefs=prefs,
                         **kwargs)

        if self.initialization_status == ContextFlags.INITIALIZED:
            self._assign_default_port_Name()

    def _instantiate_attributes_after_function(self, context=None):
        # If owner is specified but modulation has not been specified, assign to owner's value

        super()._instantiate_attributes_after_function(context=context)
        if self.owner and self.modulation is None:
            self.modulation = self.owner.modulation
        if self.modulation is not None:
            if self.modulation not in modulation_type_keywords:
                try:
                    getattr(self.function.parameters, self.modulation)
                except:
                    raise ModulatorySignalError(f"The {MODULATION} arg for {self.name} of {self.owner.name} must be "
                                                f"the name of a modulable parameter of its function "
                                                f"({self.function.__class__.__name__}) or a {MODULATION} keyword "
                                                f"(MULTIPLICATIVE, ADDITIVE, OVERRIDE, DISABLE).")

    def _instantiate_projections(self, projections, context=None):
        """Instantiate Projections specified in PROJECTIONS entry of params arg of Port's constructor

        Specification should be an existing ModulatoryProjection, or a receiver Mechanism or Port
        Disallow any other specifications (including PathwayProjections)
        Call _instantiate_projection_from_port to assign ModulatoryProjections to .efferents

        """
       # IMPLEMENTATION NOTE: THIS SHOULD BE MOVED TO COMPOSITION ONCE THAT IS IMPLEMENTED
        for receiver_spec in projections:
            projection = self._instantiate_projection_from_port(projection_spec=type(self),
                                                                receiver=receiver_spec,
                                                                # MODIFIED 8/12/19 NEW: [JDC] - MODIFIED FEEDBACK
                                                                # feedback=True,
                                                                feedback=MAYBE,
                                                                # MODIFIED 8/12/19 END
                                                                context=context)
            # Projection might be None if it was duplicate
            if projection:
                projection._assign_default_projection_name(port=self)

    def _assign_default_port_Name(self, context=None):

        # If the name is not a default name for the class,
        #    or the ModulatorySignal has no projections (which are used to name it)
        #    then return
        if (
            (
                not (
                    self.name is self.__class__.__name__
                    or self.__class__.__name__ + '-' in self.name
                )
                or len(self.efferents) == 0
            )
            and self.name not in [p.receiver.name for p in self.efferents]
        ):
            return self.name

        # Construct default name
        receiver_names = []
        receiver_owner_names = []
        receiver_owner_receiver_names = []
        class_name = self.__class__.__name__

        for projection in self.efferents:
            receiver = projection.receiver
            receiver_name = receiver.name
            receiver_owner_name = receiver.owner.name
            receiver_names.append(receiver_name)
            receiver_owner_names.append(receiver_owner_name)
            receiver_owner_receiver_names.append("{}[{}]".format(receiver_owner_name, receiver_name))

        # Only one ModulatoryProjection: "<target mech> <Port.name> <ModulatorySignal>"
        # (e.g., "Decision drift_rate ControlSignal", or "Input Layer InputPort-0 GatingSignal")
        if len(receiver_owner_receiver_names) == 1:
            default_name = receiver_owner_receiver_names[0] + " " + class_name

        # Multiple ModulatoryProjections all for same mech: "<target mech> (<Port.name>,...) <ModulatorySignal>"
        # (e.g., "Decision (drift_rate, threshold) ControlSignal" or
        #        "InputLayer (InputPort-0, InputPort-0) ControlSignal")
        elif all(name is receiver_owner_names[0] for name in receiver_owner_names):
            default_name = "{}[{}] {}".format(receiver_owner_names[0], ", ".join(receiver_names), class_name)

        # Mult ModulatoryProjections for diff mechs: "<owner mech> divergent <ModulatorySignal>"
        # (e.g., "EVC divergent ControlSignal", or "GatingMechanism divergent GatingSignal")
        else:
            default_name = self.name + " divergent " + class_name

        self.name = default_name

        return self.name



#
# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
#
# *****************************************  USER-DEFINED FUNCTION  ****************************************************

import ctypes
import numpy as np
import typecheck as tc

from inspect import signature, _empty

from psyneulink.core.components.component import ComponentError
from psyneulink.core.components.functions.function import FunctionError, Function_Base
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.keywords import \
    ADDITIVE_PARAM, CONTEXT, CUSTOM_FUNCTION, EXECUTION_ID, MULTIPLICATIVE_PARAM, OWNER, PARAMS, \
    PARAMETER_PORT_PARAMS, SELF, USER_DEFINED_FUNCTION, USER_DEFINED_FUNCTION_TYPE
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences import is_pref_set
from psyneulink.core.globals.utilities import iscompatible

from psyneulink.core import llvm as pnlvm

__all__ = ['UserDefinedFunction']

class UserDefinedFunction(Function_Base):
    """UserDefinedFunction(  \
    custom_function=None,    \
    default_variable=None,   \
    params=None,             \
    owner=None,              \
    name=None,               \
    prefs=None)

    .. _UDF_Description:

    A UserDefinedFunction (UDF) is used to "wrap" a Python function or method, including a lamdba function,
    as a PsyNeuLink `Function <Function>`, so that it can be used as the `function <Component.function>` of a
    `Component <Component>`.  This is done automatically if a Python function or method is assigned as the `function
    <Component.function>` attribute of a Component.  A Python function or method can also be wrapped explicitly,
    using the UserDefinedFunction constructor, and assigning the Python function or method to its **custom_function**
    argument.  A Python function or method wrapped as a UDF must obey the following conventions to be treated
    correctly:

    .. _UDF_Variable:

    * It must have **at least one argument** (that can be a positional or a keyword argument);  this will be treated
      as the `variable <UserDefinedFunction.variable>` attribute of the UDF's `function <UserDefinedFunction.function>`.
      When the UDF calls the function or method that it wraps, an initial attempt is made to do so with **variable**
      as the name of the first argument; if that fails, it is called positionally.  The argument is always passed as a
      2d np.array, that may contain one or more items (elements in axis 0), depending upon the Component to which the
      UDF is assigned.  It is the user's responsibility to insure that the number of items expected in the first
      argument of the function or method is compatible with the circumstances in which it will be called.
    ..
    .. _UDF_Additional_Arguments:

    * It may have have **any number of additional arguments** (positional and/or keyword);  these are treated as
      parameters of the UDF, and can be modulated by `ModulatorySignals <ModulatorySignal>` like the parameters of
      ordinary PsyNeuLink `Functions <Function>`.  If the UDF is assigned to (or automatically created for) a
      `Mechanism <Mechanism>` or `Projection <Projection>`, these parameters are each automatically assigned a
      `ParameterPort` so that they can be modulated by `ControlSignals <ControlSignal>` or `LearningSignals
      <LearningSignal>`, respectively.  If the UDF is assigned to (or automatically created for) an `InputPort` or
      `OutputPort`, and any of the parameters are specified as `Function_Modulatory_Params` (see `below
      <UDF_Modulatory_Params>`), then they can be modulated by `GatingSignals <GatingSignal>`. The function or method
      wrapped by the UDF is called with these parameters by their name and with their current values (i.e.,
      as determined by any `ModulatorySignals <ModulatorySignal>` assigned to them).
    ..
    .. _UDF_Params_Context:

    * It may include **self**, **owner**, **context**, and/or **params** arguments;  none of these are
      required, but can be included to gain access to the standard `Function` parameters (such as the history of its
      parameter values), those of the `Component` to which it has been assigned (i.e., its `owner <Function.owner>`,
      and/or receive information about the current conditions of execution.   When the function or method is called,
      an initial attempt is made to pass these arguments; if that fails, it is called again without them.
    ..
    .. _UDF_Modulatory_Params:

    * The parameters of a UDF can be specified as `Function_Modulatory_Params` in a `parameter specification dictionary
      <ParameterPort_Specification>` assigned to the **params** argument of the constructor for either the Python
      function or method, or of an explicitly defined UDF (see `examples below <UDF_Modulatory_Params_Examples>`).
      It can include either or both of the following two entries:
         *MULTIPLICATIVE_PARAM*: <parameter name>\n
         *ADDITIVE_PARAM*: <parameter name>
      These are used only when the UDF is assigned as the `function <Port_Base.function>` of an InputPort or
      OutputPort that receives one more more `GatingProjections <GatingProjection>`.

      COMMENT:
      # IMPLEMENT INTERFACE FOR OTHER MODULATION TYPES (i.e., for ability to add new custom ones)
      COMMENT

    .. tip::
       The format of the `variable <UserDefinedFunction.variable>` passed to the `custom_function
       <UserDefinedFunction.custom_function>` function can be verified by adding a ``print(variable)`` or
       ``print(type(variable))`` statement to the function.

    Examples
    --------

    **Assigning a custom function to a Mechanism**

    .. _UDF_Lambda_Function_Examples:

    The following example assigns a simple lambda function that returns the sum of the elements of a 1d array) to a
    `TransferMechanism`::

        >>> import psyneulink as pnl
        >>> my_mech = pnl.ProcessingMechanism(default_variable=[[0,0,0]],
        ...                                   function=lambda x:sum(x[0]))
        >>> my_mech.execute(input = [1, 2, 3])
        array([[6]])

    Note that the function treats its argument, x, as a 2d array, and accesses its first item for the calculation.
    This is because  the `variable <Mechanism_Base.variable>` of ``my_mech`` is defined in the **size** argument of
    its constructor as having a single item (a 1d array of length 3;  (see `size <Component.size>`).  In the
    following example, a function is defined for a Mechanism in which the variable has two items, that are summed by
    the function::

        >>> my_mech = pnl.ProcessingMechanism(default_variable=[[0],[0]],
        ...                                   function=lambda x: x[0] + x[1])
        >>> my_mech.execute(input = [[1],[2]])
        array([[3]])

    .. _UDF_Defined_Function_Examples:

    The **function** argument can also be assigned a function defined in Python::

        >>> def my_fct(variable):
        ...     return variable[0] + variable[1]
        >>> my_mech = pnl.ProcessingMechanism(default_variable=[[0],[0]],
        ...                                   function=my_fct)

    This will produce the same result as the last example.  This can be useful for assigning the function to more than
    one Component.

    More complicated functions, including ones with more than one parameter can also be used;  for example::

        >>> def my_sinusoidal_fct(input=[[0],[0]],
        ...                       phase=0,
        ...                       amplitude=1):
        ...    frequency = input[0]
        ...    t = input[1]
        ...    return amplitude * np.sin(2 * np.pi * frequency * t + phase)
        >>> my_wave_mech = pnl.ProcessingMechanism(default_variable=[[0],[0]],
        ...                                        function=my_sinusoidal_fct)

    Note that in this example, ``input`` is used as the name of the first argument, instead of ``variable``
    as in the examples above. The name of the first argument of a function to be "wrapped" as a UDF does not matter;
    in general it is good practice to use ``variable``, as the `variable <Component.variable>` of the Component
    to which the UDF is assigned is what is passed to the function as its first argument.  However, if it is helpful to
    name it something else, that is fine.

    Notice also that ``my_sinusoidal_fct`` takes two values in its ``input`` argument, that it assigns to the
    ``frequency`` and ``t`` variables of the function.  While  it could have been specified more compactly as a 1d array
    with two elements (i.e. [0,0]), it is specified in the example as a 2d array with two items to make it clear that
    it matches the format of the **default_variable** for the ProcessingMechanism to which it will be assigned,
    which requires it be formatted this way (since the `variable <Component.variable>` of all Components are converted
    to a 2d array).

    ``my_sinusoidal_fct`` also has two other arguments, ``phase`` and ``amplitude``.   When it is assigned to
    ``my_wave_mech``, those parameters are assigned to `ParameterPorts <ParameterPort>` of ``my_wave_mech``, which
    that be used to modify their values by `ControlSignals <ControlSignal>` (see `example below <_
    UDF_Control_Signal_Example>`).

    .. _UDF_Explicit_Creation_Examples:

    In all of the examples above, a UDF was automatically created for the functions assigned to the Mechanism.  A UDF
    can also be created explicitly, as follows:

        >>> my_sinusoidal_UDF = pnl.UserDefinedFunction(custom_function=my_sinusoidal_fct)
        >>> my_wave_mech = pnl.ProcessingMechanism(default_variable=[[0],[0]],
        ...                                        function=my_sinusoidal_UDF)

    When the UDF is created explicitly, parameters of the function can be included as arguments to its constructor,
    to assign them default values that differ from the those in the definition of the function, or for parameters
    that don't define default values.  For example::

        >>> my_sinusoidal_UDF = pnl.UserDefinedFunction(custom_function=my_sinusoidal_fct,
        ...                                  phase=10,
        ...                                  amplitude=3)
        >>> my_wave_mech = pnl.ProcessingMechanism(default_variable=[[0],[0]],
        ...                                        function=my_sinusoidal_UDF)

    assigns ``my_sinusoidal_fct`` as the `function <Mechanism_Base.function>` for ``my_mech``, but with the default
    values of its ``phase`` and ``amplitude`` parameters assigned new values.  This can be useful for assigning the
    same function to different Mechanisms with different default values.

    .. _UDF_Control_Signal_Example:

    Explicitly defining the UDF can also be used to `specify control <ControlSignal_Specification>` for parameters of
    the function, as in the following example::

        >>> my_mech = pnl.ProcessingMechanism(default_variable=[[0],[0]],
        ...                                   function=UserDefinedFunction(custom_function=my_sinusoidal_fct,
        ...                                                                amplitude=(1.0, pnl.CONTROL)))

    This specifies that the default value of the ``amplitude`` parameter of ``my_sinusoidal_fct`` be ``1.0``, but
    its value should be modulated by a `ControlSignal`.

    COMMENT:
    Note:  if a function explicitly defined in a UDF does not assign a default value to its first argument (i.e.,
    it is a positional argument), then the UDF that must define the variable, as in:

    Note:  if the function does not assign a default value to its first argument i.e., it is a positional arg),
    then if it is explicitly wrapped in a UDF that must define the variable, as in:
        xxx my_mech = pnl.ProcessingMechanism(default_variable=[[0],[0]],
        ...                                   function=UserDefinedFunction(default_variable=[[0],[0]],
        ...                                                                custom_function=my_sinusoidal_fct,
        ...                                                                amplitude=(1.0, pnl.CONTROL)))

    This is required so that the format of the variable can be checked for compatibilty with other Components
    with which it interacts.

    .. note::
       Built-in Python functions and methods (including numpy functions) cannot be assigned to a UDF

    COMMENT

    Custom functions can be as elaborate as desired, and can even include other PsyNeuLink `Functions <Function>`
    indirectly, such as::

        >>> import psyneulink as pnl
        >>> L = pnl.Logistic(gain = 2)
        >>> def my_fct(variable):
        ...     return L(variable) + 2
        >>> my_mech = pnl.ProcessingMechanism(size = 3, function = my_fct)
        >>> my_mech.execute(input = [1, 2, 3])  #doctest: +SKIP
        array([[2.88079708, 2.98201379, 2.99752738]])


    .. _UDF_Assign_to_Port_Examples:

    **Assigning of a custom function to a Port**

    A custom function can also be assigned as the `function <Port_Base.function>` of an `InputPort` or `OutputPort`.
    For example, the following assigns ``my_sinusoidal_fct`` to the `function <OutputPort.function>` of an OutputPort
    of ``my_mech``, rather the Mechanism's `function <Mechanism_Base.function>`::

        >>> my_wave_mech = pnl.ProcessingMechanism(size=1,
        ...                                        function=pnl.Linear,
        ...                                        output_ports=[{pnl.NAME: 'SINUSOIDAL OUTPUT',
        ...                                                       pnl.VARIABLE: [(pnl.OWNER_VALUE, 0),pnl.EXECUTION_COUNT],
        ...                                                       pnl.FUNCTION: my_sinusoidal_fct}])

    For details on how to specify a function of an OutputPort, see `OutputPort Customization <OutputPort_Customization>`.
    Below is an example plot of the output of the 'SINUSOIDAL OUTPUT' `OutputPort` from my_wave_mech above, as the
    execution count increments, when the input to the mechanism is 0.005 for 1000 runs::

.. figure:: _static/sinusoid_005.png
   :alt: Sinusoid function
   :scale: 50 %

.. _UDF_Modulatory_Params_Examples:

    The parameters of a custom function assigned to an InputPort or OutputPort can also be used for `gating
    <GatingMechanism_Specifying_Gating>`.  However, this requires that its `Function_Modulatory_Params` be specified
    (see `above <UDF_Modulatory_Params>`). This can be done by including a **params** argument in the definition of
    the function itself::

        >>> def my_sinusoidal_fct(input=[[0],[0]],
        ...                      phase=0,
        ...                      amplitude=1,
        ...                      params={pnl.ADDITIVE_PARAM:'phase',
        ...                              pnl.MULTIPLICATIVE_PARAM:'amplitude'}):
        ...    frequency = input[0]
        ...    t = input[1]
        ...    return amplitude * np.sin(2 * np.pi * frequency * t + phase)

    or in the explicit creation of a UDF::

        >>> my_sinusoidal_UDF = pnl.UserDefinedFunction(custom_function=my_sinusoidal_fct,
        ...                                             phase=0,
        ...                                             amplitude=1,
        ...                                             params={pnl.ADDITIVE_PARAM:'phase',
        ...                                                     pnl.MULTIPLICATIVE_PARAM:'amplitude'})


    The ``phase`` and ``amplitude`` parameters of ``my_sinusoidal_fct`` can now be used as the
    `Function_Modulatory_Params` for gating any InputPort or OutputPort to which the function is assigned (see
    `GatingMechanism_Specifying_Gating` and `GatingSignal_Examples`).

    **Class Definition:**


    Arguments
    ---------

    COMMENT:
        CW 1/26/18: Again, commented here is the old version, because I'm afraid I may have missed some functionality.
        custom_function : function
        specifies function to "wrap." It can be any function, take any arguments (including standard ones,
        such as :keyword:`params` and :keyword:`context`) and return any value(s), so long as these are consistent
        with the context in which the UserDefinedFunction will be used.
    COMMENT
    custom_function : function
        specifies the function to "wrap." It can be any function or method, including a lambda function;
        see `above <UDF_Description>` for additional details.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the function.
        This can be used to define an `additive_param <UserDefinedFunction.additive_param>` and/or
        `multiplicative_param <UserDefinedFunction.multiplicative_param>` for the UDF, by including one or both
        of the following entries:\n
          *ADDITIVE_PARAM*: <param_name>\n
          *MULTIPLICATIVE_PARAM*: <param_name>\n
        Values specified for parameters in the dictionary override any assigned to those parameters in arguments of
        the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).

    Attributes
    ----------

    variable: value
        format and default value of the function "wrapped" by the UDF.

    custom_function : function
        the user-specified function: called by the Function's `owner <Function_Base.owner>` when it is executed.

    additive_param : str
        this contains the name of the additive_param, if one has been specified for the UDF
        (see `above <UDF_Modulatory_Params>` for details).

    multiplicative_param : str
        this contains the name of the multiplicative_param, if one has been specified for the UDF
        (see `above <UDF_Modulatory_Params>` for details).

    COMMENT:
    enable_output_type_conversion : Bool : False
        specifies whether `function output type conversion <Function_Output_Type_Conversion>` is enabled.

    output_type : FunctionOutputType : None
        used to specify the return type for the `function <Function_Base.function>`;  `functionOuputTypeConversion`
        must be enabled and implemented for the class (see `FunctionOutputType <Function_Output_Type_Conversion>`
        for details).
    COMMENT

    owner : Component
        `component <Component>` to which the Function has been assigned.

    name : str
        the name of the Function; if it is not specified in the **name** argument of the constructor, a default is
        assigned by FunctionRegistry (see `Registry_Naming` for conventions used for default and duplicate names).

    prefs : PreferenceSet or specification dict
        the `PreferenceSet` for the Function; if it is not specified in the **prefs** argument of the
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see `Preferences`
        for details).
    """

    componentName = USER_DEFINED_FUNCTION
    componentType = USER_DEFINED_FUNCTION_TYPE

    class Parameters(Function_Base.Parameters):
        """
            Attributes
            ----------

                custom_function
                    see `custom_function <UserDefinedFunction.custom_function>`

                    :default value: None
                    :type:
        """
        custom_function = Parameter(
            None,
            stateful=False,
            loggable=False,
        )

    @tc.typecheck
    def __init__(self,
                 custom_function=None,
                 default_variable=None,
                 params=None,
                 owner=None,
                 prefs: tc.optional(is_pref_set) = None,
                 **kwargs):

        def get_cust_fct_args(custom_function):
            """Get args of custom_function
            Return:
                - value of first arg (to be used as default_variable for UDF)
                - dict with all others (to be assigned as params of UDF)
                - dict with default values (from function definition, else set to None)
            """
            try:
                arg_names = custom_function.__code__.co_varnames
            except AttributeError:
                raise FunctionError("Can't get __code__ for custom_function")
            args = {}
            defaults = {}
            for arg_name, arg in signature(custom_function).parameters.items():

                # MODIFIED 3/6/19 NEW: [JDC]
                # Custom function specified owner as arg
                if arg_name in {SELF, OWNER, CONTEXT}:
                    # Flag for inclusion in call to function
                    if arg_name == SELF:
                        self.self_arg = True
                    elif arg_name == OWNER:
                        self.owner_arg = True
                    else:
                        self.context_arg = True
                    # Skip rest, as these don't need to be params
                    continue
                # MODIFIED 3/6/19 END

                # Use definition from the function as default;
                #    this allows UDF to assign a value for this instance (including a MODULATORY spec)
                #    while assigning an actual value to current/defaults
                if arg.default is _empty:
                    defaults[arg_name] = None

                else:
                    defaults[arg_name] = arg.default

                # If arg is specified in the constructor for the UDF, assign that as its value
                if arg_name in kwargs:
                    args[arg_name] = kwargs[arg_name]
                # Otherwise, use the default value from the definition of the function
                else:
                    args[arg_name] = defaults[arg_name]

            # Assign default value of first arg as variable and remove from dict
            variable = args[arg_names[0]]
            if variable is _empty:
                variable = None
            del args[arg_names[0]]

            return variable, args, defaults

        self.self_arg = False
        self.owner_arg = False
        self.context_arg = False

        # Get variable and names of other any other args for custom_function and assign to cust_fct_params
        if params is not None and CUSTOM_FUNCTION in params:
            custom_function = params[CUSTOM_FUNCTION]
        try:
            cust_fct_variable, self.cust_fct_params, defaults = get_cust_fct_args(custom_function)
        except FunctionError:
            raise FunctionError("Assignment of a built-in function or method ({}) to a {} is not supported".
                                format(custom_function, self.__class__.__name__))

        # If params is specified as arg in custom function's definition, move it to params in UDF's constructor
        if PARAMS in self.cust_fct_params:
            if self.cust_fct_params[PARAMS]:
                if params:
                    params.update(self.cust_fct_params)
                else:
                    params = self.cust_fct_params[PARAMS]
            del self.cust_fct_params[PARAMS]

        # If context is specified as arg in custom function's definition, delete it
        if CONTEXT in self.cust_fct_params:
            if self.cust_fct_params[CONTEXT]:
                context = self.cust_fct_params[CONTEXT]
            del self.cust_fct_params[CONTEXT]

        # Assign variable to default_variable if default_variable was not specified
        if default_variable is None:
            default_variable = cust_fct_variable
        elif cust_fct_variable and not iscompatible(default_variable, cust_fct_variable):
            owner_name = ' ({})'.format(owner.name) if owner else ''
            cust_fct_name = repr(custom_function.__name__)
            raise FunctionError("Value passed as \'default_variable\' for {} {} ({}) conflicts with specification of "
                                "first argument in constructor for {} itself ({}). "
                                "Try modifying specification of \'default_variable\' "
                                "for object to which {} is being assigned{}, and/or insuring that "
                                "the first argument of {} is at least a 2d array".
                                format(self.__class__.__name__, cust_fct_name, default_variable,
                                       cust_fct_name, cust_fct_variable, cust_fct_name, owner_name, cust_fct_name))

        super().__init__(
            default_variable=default_variable,
            custom_function=custom_function,
            params=params,
            owner=owner,
            prefs=prefs,
            **self.cust_fct_params
        )

    def _handle_illegal_kwargs(self, **kwargs):
        super()._handle_illegal_kwargs(
            **{k: kwargs[k] for k in kwargs if k not in self.cust_fct_params}
        )

    def _validate_params(self, request_set, target_set=None, context=None):
        pass

    def _instantiate_attributes_before_function(self, function=None, context=None):
        super()._instantiate_attributes_before_function(function=function, context=context)
        # create transient Parameters objects for custom function params
        # done here because they need to be present before _instantiate_value which calls self.function
        for param_name in self.cust_fct_params:
            p = Parameter(self.cust_fct_params[param_name], modulable=True)
            setattr(self.parameters, param_name, p)

            p._set(p.default_value, context, skip_history=True)

    def _function(self, variable, context=None, **kwargs):

        # Update value of parms in cust_fct_params
        for param in self.cust_fct_params:

            # First check for value passed in params as runtime param:
            if PARAMS in kwargs and kwargs[PARAMS] is not None and param in kwargs[PARAMS]:
                self.cust_fct_params[param] = kwargs[PARAMS][param]
            else:
                # Otherwise, get current value from ParameterPort (in case it is being modulated by ControlSignal(s)
                self.cust_fct_params[param] = self._get_current_function_param(param, context)

        call_params = self.cust_fct_params.copy()

        # # MODIFIED 3/6/19 NEW: [JDC]
        # Add any of these that were included in the definition of the custom function:
        if self.self_arg:
            call_params[SELF] = self
        if self.owner_arg:
            call_params[OWNER] = self.owner
        if self.context_arg:
            call_params[CONTEXT] = context
        # MODIFIED 3/6/19 END

        kwargs.update(call_params)

        try:
            # Try calling with full list of args (including context and params)
            value = self.custom_function(variable, **kwargs)
        except TypeError:
            # Try calling with just variable and cust_fct_params
            value = self.custom_function(variable, **call_params)

        return self.convert_output_type(value)

    def _gen_llvm_function_body(self, ctx, builder, params, state,
                                arg_in, arg_out, *, tags:frozenset):

        # Instantiate needed ctypes
        arg_in_ct = pnlvm._convert_llvm_ir_to_ctype(arg_in.type.pointee)
        params_ct = pnlvm._convert_llvm_ir_to_ctype(params.type.pointee)
        state_ct = pnlvm._convert_llvm_ir_to_ctype(state.type.pointee)
        arg_out_ct = pnlvm._convert_llvm_ir_to_ctype(arg_out.type.pointee)
        wrapper_ct = ctypes.CFUNCTYPE(None,
                                      ctypes.POINTER(params_ct),
                                      ctypes.POINTER(state_ct),
                                      ctypes.POINTER(arg_in_ct),
                                      ctypes.POINTER(arg_out_ct))

        # we don't support passing any stateful params
        for i, p in enumerate(self.llvm_state_ids):
            assert p not in self.cust_fct_params

        def _carr_to_list(carr):
            try:
                return [_carr_to_list(x) for x in carr]
            except TypeError:
                return carr

        def _assign_to_carr(carr, vals):
            assert len(carr) == len(vals)
            for i, x in enumerate(vals):
                try:
                    carr[i] = x
                except TypeError:
                    _assign_to_carr(carr[i], x)

        def _wrapper(params, state, arg_in, arg_out):
            variable = _carr_to_list(arg_in.contents)

            llvm_params = {}
            for i, p in enumerate(self.llvm_param_ids):
                if p in self.cust_fct_params:
                    field_name = params.contents._fields_[i][0]
                    val = getattr(params.contents, field_name)
                    llvm_params[p] = val

            if self.context_arg:
                # FIXME: We can't get the context
                #        and do not support runtime params
                llvm_params[CONTEXT] = None
                llvm_params[PARAMS] = None

            value = self.custom_function(np.asfarray(variable), **llvm_params)
            _assign_to_carr(arg_out.contents, np.atleast_2d(value))

        self.__wrapper_f = wrapper_ct(_wrapper)
        # To get the right callback pointer, we need to cast to void*
        wrapper_address = ctypes.cast(self.__wrapper_f, ctypes.c_void_p)
        # Direct pointer constants don't work
        wrapper_ptr = builder.inttoptr(pnlvm.ir.IntType(64)(wrapper_address.value), builder.function.type)
        builder.call(wrapper_ptr, [params, state, arg_in, arg_out])

        return builder



from . import contrastivehebbianmechanism
from . import kohonenmechanism
from . import kwtamechanism
from . import lcamechanism
from . import recurrenttransfermechanism

from .contrastivehebbianmechanism import *
from .kohonenmechanism import *
from .kwtamechanism import *
from .lcamechanism import *
from .recurrenttransfermechanism import *

__all__ = list(kwtamechanism.__all__)
__all__.extend(lcamechanism.__all__)
__all__.extend(recurrenttransfermechanism.__all__)
__all__.extend(contrastivehebbianmechanism.__all__)
__all__.extend(kohonenmechanism.__all__)



import numpy as np
import pytest
import psyneulink as pnl


@pytest.mark.composition
@pytest.mark.benchmark(group="Gating")
@pytest.mark.parametrize("mode", ['Python',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.param('LLVMExec', marks=pytest.mark.llvm),
                                  pytest.param('LLVMRun', marks=pytest.mark.llvm),
                                  pytest.param('PTXExec', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                  pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda])
                                  ])
def test_gating(benchmark, mode):

    Input_Layer = pnl.TransferMechanism(
        name='Input_Layer',
        default_variable=np.zeros((2,)),
        function=pnl.Logistic()
    )

    Output_Layer = pnl.TransferMechanism(
        name='Output_Layer',
        default_variable=[0, 0, 0],
        function=pnl.Linear(),
        output_ports={
            pnl.NAME: 'RESULTS USING UDF',
            pnl.FUNCTION: pnl.Linear(slope=pnl.GATING)
        }
    )

    Gating_Mechanism = pnl.GatingMechanism(
        size=[1],
        gating_signals=[Output_Layer.output_port]
    )

    p_pathway = [Input_Layer, Output_Layer]

    stim_list = {
        Input_Layer: [[-1, 30], [-1, 30], [-1, 30], [-1, 30]],
        Gating_Mechanism: [[0.0], [0.5], [1.0], [2.0]]
    }

    comp = pnl.Composition(name="comp")
    comp.add_linear_processing_pathway(p_pathway)
    comp.add_node(Gating_Mechanism)

    comp.run(num_trials=4, inputs=stim_list, bin_execute=mode)

    expected_results = [
        [np.array([0., 0., 0.])],
        [np.array([0.63447071, 0.63447071, 0.63447071])],
        [np.array([1.26894142, 1.26894142, 1.26894142])],
        [np.array([2.53788284, 2.53788284, 2.53788284])]
    ]

    np.testing.assert_allclose(comp.results, expected_results)
    benchmark(comp.run, num_trials=4, inputs=stim_list, bin_execute=mode)

# DEPRECATED FUNCTIONALITY 9/26/19
# @pytest.mark.composition
# @pytest.mark.benchmark(group="Gating")
# @pytest.mark.parametrize("mode", ['Python',
#                                   pytest.param('LLVM', marks=pytest.mark.llvm),
#                                   pytest.param('LLVMExec', marks=pytest.mark.llvm),
#                                   pytest.param('LLVMRun', marks=pytest.mark.llvm),
#                                   pytest.param('PTXExec', marks=[pytest.mark.llvm, pytest.mark.cuda]),
#                                   pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda])
#                                   ])
# def test_gating_using_ControlMechanism(benchmark, mode):
#
#     Input_Layer = pnl.TransferMechanism(
#         name='Input_Layer',
#         default_variable=np.zeros((2,)),
#         function=pnl.Logistic()
#     )
#
#     Output_Layer = pnl.TransferMechanism(
#         name='Output_Layer',
#         default_variable=[0, 0, 0],
#         function=pnl.Linear(),
#         output_ports={
#             pnl.NAME: 'RESULTS USING UDF',
#             pnl.FUNCTION: pnl.Linear(slope=pnl.GATING)
#         }
#     )
#
#     Gating_Mechanism = pnl.ControlMechanism(
#         size=[1],
#         control_signals=[Output_Layer.output_port]
#     )
#
#     p_pathway = [Input_Layer, Output_Layer]
#
#     stim_list = {
#         Input_Layer: [[-1, 30], [-1, 30], [-1, 30], [-1, 30]],
#         Gating_Mechanism: [[0.0], [0.5], [1.0], [2.0]]
#     }
#
#     comp = pnl.Composition(name="comp")
#     comp.add_linear_processing_pathway(p_pathway)
#     comp.add_node(Gating_Mechanism)
#
#     comp.run(num_trials=4, inputs=stim_list, bin_execute=mode)
#
#     expected_results = [
#         [np.array([0., 0., 0.])],
#         [np.array([0.63447071, 0.63447071, 0.63447071])],
#         [np.array([1.26894142, 1.26894142, 1.26894142])],
#         [np.array([2.53788284, 2.53788284, 2.53788284])]
#     ]
#
#     np.testing.assert_allclose(comp.results, expected_results)
#     benchmark(comp.run, num_trials=4, inputs=stim_list, bin_execute=mode)



from . import function
from . import userdefinedfunction
from . import combinationfunctions
from . import interfacefunctions
from . import transferfunctions
from . import selectionfunctions
from . import statefulfunctions
from . import distributionfunctions
from . import objectivefunctions
from . import optimizationfunctions
from . import learningfunctions

from .function import *
from .userdefinedfunction import *
from .combinationfunctions import *
from .interfacefunctions import *
from .transferfunctions import *
from .selectionfunctions import *
from .statefulfunctions import *
from .distributionfunctions import *
from .objectivefunctions import *
from .optimizationfunctions import *
from .learningfunctions import *

__all__ = list(function.__all__)
__all__.extend(userdefinedfunction.__all__)
__all__.extend(combinationfunctions.__all__)
__all__.extend(interfacefunctions.__all__)
__all__.extend(transferfunctions.__all__)
__all__.extend(selectionfunctions.__all__)
__all__.extend(statefulfunctions.__all__)
__all__.extend(distributionfunctions.__all__)
__all__.extend(objectivefunctions.__all__)
__all__.extend(optimizationfunctions.__all__)
__all__.extend(learningfunctions.__all__)



import numpy as np
import pytest

from psyneulink.core.components.component import ComponentError
from psyneulink.core.components.functions.transferfunctions import Linear
from psyneulink.core.components.mechanisms.processing.transfermechanism import TransferMechanism

class TestParameterPorts:
    def test_inspect_function_params_slope_noise(self):
        A = TransferMechanism()
        B = TransferMechanism()
        assert A.function.slope == 1.0
        assert B.function.slope == 1.0
        assert A.mod_slope == [1.0]
        assert B.mod_slope == [1.0]

        assert A.noise == 0.0
        assert B.noise == 0.0
        assert A.mod_noise == 0.0
        assert B.mod_noise == 0.0

        A.function.slope = 0.2

        assert A.function.slope == 0.2
        assert B.function.slope == 1.0
        assert A.mod_slope == [1.0]
        assert B.mod_slope == [1.0]

        A.noise = 0.5

        assert A.noise == 0.5
        assert B.noise == 0.0
        assert A.mod_noise == 0.0
        assert B.mod_noise == 0.0

        B.function.slope = 0.7

        assert A.function.slope == 0.2
        assert B.function.slope == 0.7
        assert A.mod_slope == [1.0]
        assert B.mod_slope == [1.0]

        B.noise = 0.6

        assert A.noise == 0.5
        assert B.noise == 0.6
        assert A.mod_noise == 0.0
        assert B.mod_noise == 0.0

        A.execute(1.0)
        assert A.mod_slope == [0.2]

        B.execute(1.0)

        assert A.function.slope == 0.2
        assert B.function.slope == 0.7
        assert A.mod_slope == [0.2]
        assert B.mod_slope == [0.7]

        assert A.noise == 0.5
        assert B.noise == 0.6
        assert A.mod_noise == 0.5
        assert B.mod_noise == 0.6

    def test_direct_call_to_constructor_error(self):
        from psyneulink.core.components.ports.parameterport import ParameterPort, ParameterPortError
        with pytest.raises(ParameterPortError) as error_text:
            ParameterPort(owner='SOMETHING')
        assert "Contructor for ParameterPort cannot be called directly(context: None" in str(error_text.value)

class TestConfigurableParameters:
    def test_configurable_params(self):
        old_value = 0.2
        new_value = 0.7
        T = TransferMechanism(function=Linear(slope=old_value,
                                              intercept=old_value),
                              noise=old_value,
                              integration_rate=old_value)

        # SLOPE - - - - - - - -

        assert np.allclose(T.function.slope, old_value)
        assert np.allclose(T.mod_slope, old_value)

        T.function.slope = new_value

        assert np.allclose(T.function.slope, new_value)
        assert np.allclose(T.mod_slope, old_value)

        # INTERCEPT - - - - - - - -

        assert np.allclose(T.function.intercept, old_value)
        assert np.allclose(T.mod_intercept, old_value)

        T.function.intercept = new_value

        assert np.allclose(T.function.intercept, new_value)
        assert np.allclose(T.mod_intercept, old_value)

        # SMOOTHING FACTOR - - - - - - - -

        assert np.allclose(T.integration_rate, old_value)
        assert np.allclose(T.mod_integration_rate, old_value)

        T.integration_rate = new_value

        # KAM changed 3/2/18 --
        # function_params looks at ParameterPort value, so this will not update until next execution
        assert np.allclose(T.integration_rate, new_value)
        assert np.allclose(T.mod_integration_rate, old_value)

        # NOISE - - - - - - - -

        assert np.allclose(T.noise, old_value)
        assert np.allclose(T.mod_noise, old_value)

        T.noise = new_value

        # KAM changed 3/2/18 --
        # function_params looks at ParameterPort value, so this will not update until next execution
        assert np.allclose(T.noise, new_value)
        assert np.allclose(T.mod_noise, old_value)

        T.execute(1.0)

        assert np.allclose(T.function.slope, new_value)
        assert np.allclose(T.mod_slope, new_value)

        assert np.allclose(T.function.intercept, new_value)
        assert np.allclose(T.mod_intercept, new_value)

        assert np.allclose(T.integration_rate, new_value)
        assert np.allclose(T.mod_integration_rate, new_value)

        assert np.allclose(T.noise, new_value)
        assert np.allclose(T.mod_noise, new_value)

class TestModParams:
    def test_mod_param_error(self):
        T = TransferMechanism()
        with pytest.raises(ComponentError) as error_text:
            T.mod_slope = 20.0
        assert "directly because it is computed by the ParameterPort" in str(error_text.value)



import numpy as np
import os
import pytest
from itertools import combinations

from psyneulink.core import llvm as pnlvm
from psyneulink.core.components.functions.transferfunctions import Linear
from psyneulink.core.components.mechanisms.processing.integratormechanism import IntegratorMechanism
from psyneulink.core.components.mechanisms.processing.transfermechanism import TransferMechanism
from psyneulink.core.compositions.composition import Composition

debug_options=["const_input=[[[7]]]", "const_input", "const_data", "const_params", "const_data", "const_state"]
options_combinations = (";".join(("debug_info", *c)) for i in range(len(debug_options) + 1) for c in combinations(debug_options, i))

@pytest.mark.composition
@pytest.mark.parametrize("mode", [
                                  pytest.param('LLVMRun', marks=pytest.mark.llvm),
                                  pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda])
                                  ])
@pytest.mark.parametrize("debug_env", [comb for comb in options_combinations if comb.count("const_input") < 2])
def test_debug_comp(mode, debug_env):
    # save old debug env var
    old_env = os.environ.get("PNL_LLVM_DEBUG")
    if debug_env is not None:
        os.environ["PNL_LLVM_DEBUG"] = debug_env
        pnlvm.debug._update()

    comp = Composition()
    A = IntegratorMechanism(default_variable=1.0, function=Linear(slope=5.0))
    B = TransferMechanism(function=Linear(slope=5.0), integrator_mode=True)
    comp.add_linear_processing_pathway([A, B])

    inputs_dict = {A: [5]}
    output1 = comp.run(inputs=inputs_dict, bin_execute=mode)
    output2 = comp.run(inputs=inputs_dict, bin_execute=mode)
    # restore old debug env var and cleanup the debug configuration
    if old_env is None:
        del os.environ["PNL_LLVM_DEBUG"]
    else:
        os.environ["PNL_LLVM_DEBUG"] = old_env
    pnlvm.debug._update()

    assert len(comp.results) == 2

    if "const_input=" in debug_env:
        expected1 = 87.5
        expected2 = 131.25
    elif "const_input" in debug_env:
        expected1 = 12.5
        expected2 = 18.75
    else:
        expected1 = 62.5
        expected2 = 93.75

    if "const_state" in debug_env:
        expected2 = expected1


    assert np.allclose(expected1, output1[0][0])
    assert np.allclose(expected2, output2[0][0])
